class BaseTypes:
    BOX = "BOX"
    TUBULAR = "TUB"
    IPROFILE = "I"
    TPROFILE = "T"
    ANGULAR = "HP"
    CHANNEL = "UNP"
    CIRCULAR = "CIRC"
    GENERAL = "GENERAL"
    FLATBAR = "FB"


class SectionCat:

    BASETYPES = BaseTypes

    box = [BASETYPES.BOX, "BG", "CG"]
    shs = ["SHS"]
    rhs = ["RHS", "URHS"]
    tubular = [BASETYPES.TUBULAR, "PIPE", "OD"]
    iprofiles = ["HEA", "HEB", "HEM", "IPE"]
    igirders = [BASETYPES.IPROFILE, "IG"]
    tprofiles = [BASETYPES.TPROFILE, "TG"]
    angular = [BASETYPES.ANGULAR]
    channels = [BASETYPES.CHANNEL]
    circular = [BASETYPES.CIRCULAR]
    general = [BASETYPES.GENERAL, "GENBEAM"]
    flatbar = [BASETYPES.FLATBAR]

    @classmethod
    def isbeam(cls, bmtype):
        for key, val in cls.__dict__.items():
            if bmtype in val:
                return True
        return False

    @staticmethod
    def _get_sec_type(section_ref):
        from ada import Beam, Section

        if type(section_ref) is Section:
            return section_ref.type.upper()
        if type(section_ref) is Beam:
            return section_ref.section.type.upper()
        else:
            return section_ref.upper()

    @classmethod
    def get_shape_type(cls, bmtype):
        type_map = [
            (cls.is_i_profile, cls.BASETYPES.IPROFILE),
            (cls.is_angular, cls.BASETYPES.ANGULAR),
            (cls.is_flatbar, cls.BASETYPES.FLATBAR),
            (cls.is_box_profile, cls.BASETYPES.BOX),
            (cls.is_t_profile, cls.BASETYPES.TPROFILE),
            (cls.is_channel_profile, cls.BASETYPES.CHANNEL),
            (cls.is_tubular_profile, cls.BASETYPES.TUBULAR),
            (cls.is_circular_profile, cls.BASETYPES.CIRCULAR),
            (cls.is_general, cls.BASETYPES.GENERAL),
        ]

        for type_func, return_type in type_map:
            if type_func(bmtype):
                return return_type

    @classmethod
    def is_i_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.igirders + cls.iprofiles else False

    @classmethod
    def is_t_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.tprofiles else False

    @classmethod
    def is_box_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.box + cls.shs + cls.rhs else False

    @classmethod
    def is_circular_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.circular else False

    @classmethod
    def is_tubular_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.tubular else False

    @classmethod
    def is_channel_profile(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.channels else False

    @classmethod
    def is_flatbar(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.flatbar else False

    @classmethod
    def is_general(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.general else False

    @classmethod
    def is_angular(cls, bmtype):
        return True if cls._get_sec_type(bmtype) in cls.angular else False

    @classmethod
    def is_strong_axis_symmetric(cls, section):
        """

        :param section:
        :type section: ada.Section
        :return:
        """
        return section.w_top == section.w_btn and section.t_ftop == section.t_fbtn



import json
import logging

import h5py

from ada.concepts.containers import Beams, Materials, Nodes, Sections
from ada.concepts.levels import FEM, Assembly, Part
from ada.concepts.points import Node
from ada.concepts.structural import Beam, Material, Section
from ada.fem import Elem
from ada.fem.containers import FemElements
from ada.materials.metals import CarbonSteel

from .utils import from_safe_name, str_fix


def read_assembly_from_cache(h5_filename, assembly=None):
    import h5py

    f = h5py.File(h5_filename, "r")
    info = f["INFO"].attrs
    a = Assembly(info["NAME"]) if assembly is None else assembly
    walk_parts(f.get("PARTS"), a)

    return a


def walk_parts(cache_p, parent):
    for name, p in cache_p.items():
        if type(p) is h5py.Dataset:
            continue
        unsafe_name = from_safe_name(name)
        parent_name = from_safe_name(p.attrs.get("PARENT", ""))
        if parent.name != parent_name:
            logging.error("Unable to retrieve proper Hierarchy from HDF5 cache")
        curr_p = parent.add_part(get_part_from_cache(unsafe_name, p))
        walk_parts(p, curr_p)


def get_part_from_cache(name, part_cache):
    meta_str = part_cache.attrs.get("METADATA")
    metadata = None
    if meta_str is not None:
        metadata = json.loads(meta_str)

    p = Part(name, metadata=metadata)
    fem = part_cache.get("FEM")
    if fem is not None:
        p._fem = get_fem_from_cache(fem)

    node_group = part_cache.get("NODES")
    if node_group is not None:
        p._nodes = get_nodes_from_cache(node_group, p)

    sections = get_sections_from_cache(part_cache, p)
    if sections is not None:
        p._sections = sections

    materials = get_materials_from_cache(part_cache, p)
    if materials is not None:
        p._materials = materials

    beams = get_beams_from_cache(part_cache, p)
    if beams is not None:
        p._beams = beams

    return p


def get_beams_from_cache(part_cache, parent: Part):
    prefix = "BEAMS"
    beams_str = part_cache.get(f"{prefix}_STR")
    beams_int = part_cache.get(f"{prefix}_INT")
    beams_up = part_cache.get(f"{prefix}_UP")
    if beams_str is None:
        return None

    def bm_from_cache(bm_str, bm_int, bm_up):
        nid1, nid2 = [parent.nodes.from_id(nid) for nid in bm_int]
        guid, name, sec_name, mat_name, meta_str = str_fix(bm_str)
        sec = parent.sections.get_by_name(sec_name)
        mat = parent.materials.get_by_name(mat_name)
        metadata = None
        if meta_str is not None:
            metadata = json.loads(meta_str)
        return Beam(name, nid1, nid2, sec=sec, mat=mat, guid=guid, parent=parent, metadata=metadata, up=bm_up)

    bm_zip = zip(beams_str, beams_int, beams_up)

    return Beams([bm_from_cache(bm_str, bm_int, bm_up) for bm_str, bm_int, bm_up in bm_zip], parent=parent)


def get_sections_from_cache(part_cache, parent):
    sections_str = part_cache.get("SECTIONS_STR")
    sections_int = part_cache.get("SECTIONS_INT")
    if sections_str is None:
        return None

    def sec_from_list(sec_str, sec_int):
        guid, name, units, sec_type = str_fix(sec_str)
        r, wt, h, w_top, w_btn, t_w, t_ftop, t_fbtn, sec_id = [x if x != 0 else None for x in sec_int]
        return Section(
            name=name,
            guid=guid,
            sec_id=sec_id,
            units=units,
            sec_type=sec_type,
            r=r,
            wt=wt,
            h=h,
            w_top=w_top,
            w_btn=w_btn,
            t_w=t_w,
            t_ftop=t_ftop,
            t_fbtn=t_fbtn,
        )

    return Sections(
        [sec_from_list(sec_str, sec_int) for sec_str, sec_int in zip(sections_str, sections_int)], parent=parent
    )


def get_materials_from_cache(part_cache, parent):
    mat_str = part_cache.get("MATERIALS_STR")
    mat_int = part_cache.get("MATERIALS_INT")

    if mat_str is None:
        return None

    def mat_from_list(mat_int, mat_str):
        guid, name, units = str_fix(mat_str)
        E, rho, sigy, mat_id = mat_int
        return Material(
            name=name,
            guid=guid,
            mat_id=mat_id,
            units=units,
            mat_model=CarbonSteel(E=E, rho=rho, sig_y=sigy),
            parent=parent,
        )

    return Materials([mat_from_list(mat_int, mat_str) for mat_int, mat_str in zip(mat_int, mat_str)], parent=parent)


def get_fem_from_cache(cache_fem):
    node_groups = cache_fem["NODES"]
    fem = FEM(cache_fem.attrs["NAME"])
    fem.nodes = get_nodes_from_cache(node_groups, fem)
    elements = []
    for eltype, mesh in cache_fem["MESH"].items():
        el_ids = mesh["ELEMENTS"][()]
        elements += [Elem(el_id[0], [fem.nodes.from_id(eli) for eli in el_id[1:]], eltype) for el_id in el_ids]
    fem.elements = FemElements(elements, fem)

    return fem


def get_nodes_from_cache(node_group, parent):
    points_in = node_group[()]
    points = [Node(n[1:], n[0]) for n in points_in]
    return Nodes(points, parent=parent)



from typing import Union

from ada.core.utils import bool2text
from ada.fem.steps import Step, StepEigen, StepImplicit


def step_str(step: Union[StepEigen, StepImplicit]):
    from .writer import bc_str, interactions_str, load_str

    bcstr = "\n".join([bc_str(bc) for bc in step.bcs.values()]) if len(step.bcs) > 0 else "** No BCs"
    lstr = "\n".join([load_str(l) for l in step.loads]) if len(step.loads) > 0 else "** No Loads"

    int_str = (
        "\n".join([interactions_str(interact) for interact in step.interactions.values()])
        if len(step.interactions.values()) > 0
        else "** No Interactions"
    )

    nodal = []
    elem = []
    for fi in step.field_outputs:
        nodal += fi.nodal
        elem += fi.element

    nodal_str = "*node file\n" + ", ".join(nodal) if len(nodal) > 0 else "** No nodal output"
    elem_str = "*el file\n" + ", ".join(elem) if len(elem) > 0 else "** No elem output"

    step_type_map = {
        Step.TYPES.STATIC: static_step,
        Step.TYPES.EIGEN: eigen_step,
    }

    step_str_writer = step_type_map.get(step.type, None)
    if step_str_writer is None:
        raise ValueError(f'Currently unsupported Step Type "{step.type}"')

    step_type_str = step_str_writer(step)
    return f"""**
** STEP: {step.name}
**
{step_type_str}
**
** BOUNDARY CONDITIONS
**
{bcstr}
**
** LOADS
**
{lstr}
**
** INTERACTIONS
**
{int_str}
**
** OUTPUT REQUESTS
**
{nodal_str}
{elem_str}
*End Step"""


def static_step(step: StepImplicit):
    return f"""*Step, nlgeom={bool2text(step.nl_geom)}, inc={step.total_incr}
*Static
 {step.init_incr}, {step.total_time}, {step.min_incr}, {step.max_incr}"""


def eigen_step(step: StepEigen):
    return f"""*Step, name={step.name}
*Frequency
 {step.num_eigen_modes}"""



from __future__ import annotations

from itertools import chain
from typing import List, Union

import gmsh
import numpy as np

from ada import FEM, Beam, Node, Pipe, Plate
from ada.concepts.transforms import Placement
from ada.core.utils import make_name_fem_ready
from ada.fem import Elem, FemSection, FemSet
from ada.fem.shapes import ElemType
from ada.fem.shapes.mesh_types import abaqus_to_meshio_type, gmsh_to_meshio_ordering

from .common import gmsh_map
from .concepts import GmshData


def add_fem_sections(model: gmsh.model, fem: FEM, model_obj: Union[Beam, Plate, Pipe], gmsh_data: GmshData) -> None:
    if type(model_obj) is Beam and gmsh_data.geom_repr == ElemType.SHELL:
        get_sh_sections_for_beam_obj(model, model_obj, gmsh_data, fem)
        return None

    if type(model_obj) is Pipe and gmsh_data.geom_repr == ElemType.SHELL:
        get_sh_sections_for_pipe_obj(model, model_obj, gmsh_data, fem)
        return None

    if gmsh_data.geom_repr == ElemType.SHELL:
        get_sh_sections_for_plate_obj(model, model_obj, gmsh_data, fem)
    elif gmsh_data.geom_repr == ElemType.SOLID:
        get_so_sections(model, model_obj, gmsh_data, fem)
    elif gmsh_data.geom_repr == ElemType.LINE:
        get_bm_sections(model, model_obj, gmsh_data, fem)
    else:
        raise ValueError(
            f'Unrecognized geometric representation "{gmsh_data.geom_repr}". ' f"Only {ElemType.all} are supported"
        )


def get_sh_sections_for_beam_obj(model: gmsh.model, beam: Beam, gmsh_data: GmshData, fem: FEM):
    from ada.sections.bm_sh_ident import eval_thick_normal_from_cog_of_beam_plate

    pl1 = Placement(beam.n1.p, beam.yvec, beam.up, beam.xvec)
    for _, ent in gmsh_data.entities:
        _, _, param = model.mesh.getNodes(2, ent, True)
        normal = np.array([0.0 if abs(x) == 0.0 else x for x in model.getNormal(ent, param)[:3]])
        cog = model.occ.getCenterOfMass(2, ent)
        pc = eval_thick_normal_from_cog_of_beam_plate(beam.section, cog, normal, pl1)

        _, tags, _ = model.mesh.getElements(2, ent)
        elements = [fem.elements.from_id(x) for x in chain.from_iterable(tags)]
        set_name = make_name_fem_ready(f"el{beam.name}_e{ent}_{pc.type}_sh")
        fem_sec_name = make_name_fem_ready(f"d{beam.name}_e{ent}_{pc.type}_sh")

        add_shell_section(set_name, fem_sec_name, normal, pc.thick, elements, beam, fem)


def get_sh_sections_for_pipe_obj(model: gmsh.model, model_obj: Pipe, gmsh_data: GmshData, fem: FEM):
    thickness = model_obj.section.wt
    normal = model_obj.segments[0].zvec

    for i, (_, ent) in enumerate(gmsh_data.entities):
        _, tags, _ = model.mesh.getElements(2, ent)
        set_name = make_name_fem_ready(f"el{model_obj.name}_e{ent}_pipe_sh")
        fem_sec_name = make_name_fem_ready(f"d{model_obj.name}_e{ent}_pipe_sh")
        elements = [fem.elements.from_id(x) for x in chain.from_iterable(tags)]
        add_shell_section(set_name, fem_sec_name, normal, thickness, elements, model_obj, fem)


def get_sh_sections_for_plate_obj(model: gmsh.model, model_obj: Plate, gmsh_data: GmshData, fem: FEM):
    tags = []
    for dim, ent in gmsh_data.entities:
        _, tag, _ = model.mesh.getElements(2, ent)
        tags += tag

    elements = [fem.elements.from_id(x) for x in chain.from_iterable(tags)]

    thickness = model_obj.t
    normal = model_obj.n

    set_name = make_name_fem_ready(f"el{model_obj.name}_sh")
    fem_sec_name = make_name_fem_ready(f"d{model_obj.name}_sh")

    add_shell_section(set_name, fem_sec_name, normal, thickness, elements, model_obj, fem)


def add_shell_section(
    set_name, fem_sec_name, normal, thickness, elements, model_obj: Union[Beam, Plate, Pipe], fem: FEM
):
    fem_set = FemSet(set_name, elements, FemSet.TYPES.ELSET)
    props = dict(local_z=normal, thickness=thickness, int_points=5)
    fem_sec = FemSection(fem_sec_name, ElemType.SHELL, fem_set, model_obj.material, **props)
    add_sec_to_fem(fem, fem_sec, fem_set)


def get_bm_sections(model: gmsh.model, beam: Beam, gmsh_data, fem: FEM):
    from ada.core.utils import vector_length

    tags = []
    for dim, ent in gmsh_data.entities:
        _, tag, _ = model.mesh.getElements(1, ent)
        tags += tag

    elements = [fem.elements.from_id(elid) for elid in chain.from_iterable(tags)]

    set_name = make_name_fem_ready(f"el{beam.name}_set_bm")
    fem_sec_name = make_name_fem_ready(f"d{beam.name}_sec_bm")
    fem_set = FemSet(set_name, elements, FemSet.TYPES.ELSET, parent=fem)
    fem_sec = FemSection(fem_sec_name, ElemType.LINE, fem_set, beam.material, beam.section, beam.ori[2], refs=[beam])

    add_sec_to_fem(fem, fem_sec, fem_set)
    if beam.hinge_prop is None:
        return
    end1_p = beam.hinge_prop.end1.concept_node.p if beam.hinge_prop.end1 is not None else None
    end2_p = beam.hinge_prop.end2.concept_node.p if beam.hinge_prop.end2 is not None else None
    if beam.hinge_prop is not None:
        for el in elements:
            n1 = el.nodes[0]
            n2 = el.nodes[-1]
            el.hinge_prop = beam.hinge_prop
            if beam.hinge_prop.end1 is not None and vector_length(end1_p - n1.p) == 0.0:
                el.hinge_prop.end1.fem_node = n1

            if beam.hinge_prop.end2 is not None and vector_length(end2_p - n2.p) == 0.0:
                el.hinge_prop.end2.fem_node = n2


def get_so_sections(model: gmsh.model, solid_object: Beam, gmsh_data: GmshData, fem: FEM):
    tags = []
    for dim, ent in gmsh_data.entities:
        _, tag, _ = model.mesh.getElements(3, ent)
        tags += tag

    elements = [fem.elements.from_id(elid) for elid in chain.from_iterable(tags)]

    set_name = make_name_fem_ready(f"el{solid_object.name}_so")
    fem_sec_name = make_name_fem_ready(f"d{solid_object.name}_so")

    fem_set = FemSet(set_name, elements, FemSet.TYPES.ELSET, parent=fem)
    fem_sec = FemSection(fem_sec_name, ElemType.SOLID, fem_set, solid_object.material)

    add_sec_to_fem(fem, fem_sec, fem_set)


def add_sec_to_fem(fem: FEM, fem_section: FemSection, fem_set: FemSet):
    fem_set_ = fem.sets.add(fem_set)
    fem_section.elset = fem_set_
    fem.add_section(fem_section)


def get_point(model: gmsh.model, p, tol):
    tol_vec = np.array([tol, tol, tol])
    lower = np.array(p) - tol_vec
    upper = np.array(p) + tol_vec
    return model.getEntitiesInBoundingBox(*lower.tolist(), *upper.tolist(), 0)


def get_nodes_from_gmsh(model: gmsh.model, fem: FEM) -> List[Node]:
    nodes = list(model.mesh.getNodes(-1, -1))
    node_ids = nodes[0]
    node_coords = nodes[1].reshape(len(node_ids), 3)
    return [Node(coord, nid, parent=fem) for nid, coord in zip(node_ids, node_coords)]


def get_elements_from_entity(model: gmsh.model, ent, fem: FEM, dim) -> List[Elem]:
    elem_types, elem_tags, elem_node_tags = model.mesh.getElements(dim, ent)
    elements = []
    for k, element_list in enumerate(elem_tags):
        el_name, _, _, numv, _, _ = model.mesh.getElementProperties(elem_types[k])
        if el_name == "Point":
            continue
        elem_type = gmsh_map[el_name]
        for j, eltag in enumerate(element_list):
            nodes = []
            for i in range(numv):
                idtag = numv * j + i
                p1 = elem_node_tags[k][idtag]
                nodes.append(fem.nodes.from_id(p1))

            new_nodes = node_reordering(elem_type, nodes)
            if new_nodes is not None:
                nodes = new_nodes

            el = Elem(eltag, nodes, elem_type, parent=fem)
            elements.append(el)
    return elements


def get_elements_from_entities(model: gmsh.model, entities, fem: FEM) -> List[Elem]:
    elements = []
    for dim, ent in entities:
        elements += get_elements_from_entity(model, ent, fem, dim)
    return elements


def is_reorder_necessary(elem_type):
    meshio_type = abaqus_to_meshio_type[elem_type]
    if meshio_type in gmsh_to_meshio_ordering.keys():
        return True
    else:
        return False


def node_reordering(elem_type, nodes):
    """Based on work in meshio"""
    meshio_type = abaqus_to_meshio_type[elem_type]
    order = gmsh_to_meshio_ordering.get(meshio_type, None)
    if order is None:
        return None

    return [nodes[i] for i in order]


def build_bm_lines(model: gmsh.model, bm: Beam, point_tol):
    p1, p2 = bm.n1.p, bm.n2.p

    midpoints = bm.calc_con_points(point_tol=point_tol)

    if bm.connected_end1 is not None:
        p1 = bm.connected_end1.centre
    if bm.connected_end2 is not None:
        p2 = bm.connected_end2.centre

    s = get_point(model, p1, point_tol)
    e = get_point(model, p2, point_tol)

    if len(s) > 1:
        raise ValueError("Multiple nodes")

    if len(s) == 0:
        s = add_point(model, p1.tolist())
    if len(e) == 0:
        e = add_point(model, p2.tolist())

    if len(midpoints) == 0:
        return add_line(model, s, e)

    prev_p = None
    entities = []
    for i, con_centre in enumerate(midpoints):
        midp = get_point(model, con_centre, point_tol)
        if len(midp) == 0:
            midp = add_point(model, con_centre)

        if prev_p is None:
            entities += add_line(model, s, midp)
            prev_p = midp
            continue
        entities += add_line(model, prev_p, midp)
        prev_p = midp

    if prev_p is None:
        entities += add_line(model, s, e)
    else:
        entities += add_line(model, prev_p, e)

    return entities


def add_line(model: gmsh.model, s, e):
    line = gmsh.model.geo.addLine(s[0][1], e[0][1])
    model.geo.synchronize()
    return [(1, line)]


def add_point(model: gmsh.model, p):
    point = [(0, model.geo.addPoint(*p))]
    model.geo.synchronize()
    return point



import json
import pathlib
import re
from collections import OrderedDict
from typing import Tuple, Union

import ada.core.utils
from ada.core.exceptions import UnsupportedUnits

from . import Section
from .categories import SectionCat

digit = r"\d{0,5}\.?\d{0,5}|\d{0,5}|\d{0,5}\/\d{0,5}"
flex = r"?:\.|[A-Z]|"


def profile_db_collect(sec_type: str, dim: str, units: str = "m"):
    """Return a section object based on values in a profile db json document. Source JSON is in units 'm' meters."""
    scale_map = {"mm": 1000, "m": 1.0}

    scale_factor = scale_map.get(units, None)

    if scale_factor is None:
        raise UnsupportedUnits(f'Units "{units}" is not supported')

    if sec_type == "IP":
        sec_type = "IPE"
    dir_path = pathlib.Path(__file__).resolve().parent

    with open(dir_path / "resources" / "ProfileDB.json") as data_file:
        profile_db_main = json.load(data_file, object_pairs_hook=OrderedDict)

    profile_db = profile_db_main["ProfileDB"]
    sec_dim = [int(x) for x in dim.split("x")]
    sec_name_alt1 = sec_type + str(dim)

    if "x" in str(dim):
        sec_name_alt2 = sec_type + str(sec_dim[0]) + "x" + str(sec_dim[1])
    else:
        sec_name_alt2 = "Unknown"

    if sec_type not in profile_db.keys():
        return None

    if sec_name_alt1 in profile_db[sec_type]:
        sec_name = sec_name_alt1
    elif sec_name_alt2 in profile_db[sec_type]:
        sec_name = sec_name_alt2
    else:
        return None
    h = float(profile_db[sec_type][sec_name]["Height"]) * scale_factor
    w_top = float(profile_db[sec_type][sec_name]["Width"]) * scale_factor
    w_btn = float(profile_db[sec_type][sec_name]["Width"]) * scale_factor
    t_w = float(profile_db[sec_type][sec_name]["t_w"]) * scale_factor
    t_fbtn = float(profile_db[sec_type][sec_name]["t_f"]) * scale_factor
    t_ftop = float(profile_db[sec_type][sec_name]["t_f"]) * scale_factor
    return Section(
        sec_name,
        sec_type=sec_type,
        h=h,
        w_top=w_top,
        w_btn=w_btn,
        t_w=t_w,
        t_fbtn=t_fbtn,
        t_ftop=t_ftop,
        sec_str=sec_name,
        metadata=dict(cad_str=sec_name),
        units=units,
    )


def interpret_section_str(in_str: str, s=0.001, units="m"):
    """

    :param in_str:
    :param s: Scale factor
    :param units: The desired units after applied scale factor
    :return: Two section (to account for potential beam tapering)
    """
    rdoff = ada.core.utils.roundoff

    re_in = re.IGNORECASE | re.DOTALL
    for box in SectionCat.box:
        res = re.search(
            r"({box})({flex})({digit})x({digit})x({digit})x({digit})".format(box=box, flex=flex, digit=digit),
            in_str,
            re_in,
        )
        if res is None:
            continue
        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
        tf = [rdoff(float(x) * s) for x in res.group(5).split("/")]
        sec = Section(
            in_str,
            h=h[0],
            sec_type=box,
            w_btn=width[0],
            w_top=width[0],
            t_fbtn=tf[0],
            t_ftop=tf[0],
            t_w=tw[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        if "/" in in_str:
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=box,
                w_btn=width[-1],
                w_top=width[-1],
                t_fbtn=tf[-1],
                t_ftop=tf[-1],
                t_w=tw[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
        else:
            tap = sec
        return sec, tap

    for shs in SectionCat.shs:
        res = re.search(
            r"({hol})({flex})({digit})x({digit})".format(hol=shs, flex=flex, digit=digit),
            in_str,
            re_in,
        )

        if res is None:
            continue

        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        sec = Section(
            in_str,
            h=h[0],
            sec_type=shs,
            w_btn=h[0],
            w_top=h[0],
            t_fbtn=width[0],
            t_ftop=width[0],
            t_w=width[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        if "/" in in_str:
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=shs,
                w_btn=h[-1],
                w_top=h[-1],
                t_fbtn=width[-1],
                t_ftop=width[-1],
                t_w=width[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
        else:
            tap = sec
        return sec, tap

    for rhs in SectionCat.rhs:
        res = re.search(
            r"({rhs})({flex})({digit})x({digit})x({digit})".format(rhs=rhs, flex=flex, digit=digit),
            in_str,
            re_in,
        )

        if res is None:
            continue
        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
        sec = Section(
            in_str,
            h=h[0],
            sec_type=rhs,
            w_btn=width[0],
            w_top=width[0],
            t_fbtn=tw[0],
            t_ftop=tw[0],
            t_w=tw[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        tap = Section(
            in_str + "_e",
            h=h[-1],
            sec_type=rhs,
            w_btn=width[-1],
            w_top=width[-1],
            t_fbtn=tw[-1],
            t_ftop=tw[-1],
            t_w=tw[-1],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        return sec, tap

    for ig in SectionCat.igirders:
        res = re.search(
            "({ig})({flex})({digit})x({digit})x({digit})x({digit})".format(ig=ig, flex=flex, digit=digit),
            in_str,
            re_in,
        )
        if res is None:
            continue
        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        wt = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
        tf = [rdoff(float(x) * s) for x in res.group(5).split("/")]
        sec = Section(
            in_str,
            h=h[0],
            sec_type=ig,
            w_btn=wt[0],
            w_top=wt[0],
            t_fbtn=tf[0],
            t_ftop=tf[0],
            t_w=tw[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        tap = Section(
            in_str + "_e",
            h=h[-1],
            sec_type=ig,
            w_btn=wt[-1],
            w_top=wt[-1],
            t_fbtn=tf[-1],
            t_ftop=tf[-1],
            t_w=tw[-1],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        return sec, tap

    for tg in SectionCat.tprofiles:
        res = re.search(
            "({ig})({flex})({digit})x({digit})x({digit})x({digit})".format(ig=tg, flex=flex, digit=digit),
            in_str,
            re_in,
        )
        if res is None:
            continue
        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        wt = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
        tf = [rdoff(float(x) * s) for x in res.group(5).split("/")]
        sec = Section(
            in_str,
            h=h[0],
            sec_type=tg,
            w_btn=wt[0],
            w_top=tw[0],
            t_fbtn=tf[0],
            t_ftop=tf[0],
            t_w=tw[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        tap = Section(
            in_str + "_e",
            h=h[-1],
            sec_type=tg,
            w_btn=wt[-1],
            w_top=tw[-1],
            t_fbtn=tf[-1],
            t_ftop=tf[-1],
            t_w=tw[-1],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        return sec, tap

    if "pipe" not in in_str.lower():
        for ipe in SectionCat.iprofiles:
            res = re.search("({ipe})({digit})".format(ipe=ipe, digit=digit), in_str, re_in)
            if res is not None:
                sec = profile_db_collect(ipe, res.group(2), units=units)
                return sec, sec
            else:
                if "HE" not in ipe:
                    continue
                shuffle = "({ipe})({digit}){end}".format(ipe=ipe[:2], digit=digit, end=ipe[2:])
                res = re.search(shuffle, in_str, re_in)
                if res is not None:
                    sec = profile_db_collect(ipe, res.group(2), units=units)
                    return sec, sec

    for tub in SectionCat.tubular:
        res = re.search("({tub})({digit})x({digit})".format(tub=tub, digit=digit), in_str, re_in)
        if res is None:
            continue
        fac = 0.5 if tub in ["OD", "O"] else 1.0
        r = [rdoff(float(x) * s * fac) for x in res.group(2).split("/")]
        wt = [rdoff(float(x) * s) for x in res.group(3).split("/")]
        sec = Section(
            in_str,
            sec_type=tub,
            r=r[0],
            wt=wt[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        tap = Section(
            in_str + "_e",
            sec_type=tub,
            r=r[-1],
            wt=wt[-1],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        return sec, tap

    for ang in SectionCat.angular:
        res = re.search("({ang})({digit})x({digit})".format(ang=ang, digit=digit), in_str, re_in)
        if res is None:
            continue
        sec = profile_db_collect(ang, "{}x{}".format(res.group(2), res.group(3)), units=units)
        return sec, sec

    for circ in SectionCat.circular:
        res = re.search("{circ}(.*?)$".format(circ=circ), in_str, re_in)
        if res is None:
            continue
        sec = Section(
            in_str,
            sec_type=circ,
            r=rdoff(float(res.group(1)) * s),
            metadata=dict(cad_str=in_str),
            units=units,
        )
        return sec, sec

    for cha in SectionCat.channels:
        res = re.search("({ang})({digit})".format(ang=cha, digit=digit), in_str, re_in)
        if res is None:
            continue
        sec = profile_db_collect(cha, res.group(2), units=units)
        return sec, sec

    for flat in SectionCat.flatbar:
        res = re.search("({flat})({digit})x({digit})".format(flat=flat, digit=digit), in_str, re_in)
        if res is None:
            continue
        h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
        width = [rdoff(float(x) * s) for x in res.group(3).split("/")]

        sec = Section(
            in_str,
            sec_type=flat,
            h=h[0],
            w_top=width[0],
            w_btn=width[0],
            metadata=dict(cad_str=in_str),
            units=units,
        )
        if "/" in in_str:
            tap = Section(
                in_str + "_e",
                sec_type=flat,
                h=h[-1],
                w_top=width[-1],
                w_btn=width[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
        else:
            tap = sec
        return sec, tap

    raise ValueError(f'Unable to interpret section str "{in_str}"')


def get_section(sec: Union[Section, str]) -> Tuple[Section, Section]:
    if type(sec) is Section:
        return sec, sec
    elif type(sec) is str:
        return interpret_section_str(sec)
    else:
        raise ValueError("Unable to find beam section based on input: {}".format(sec))



import pathlib
import shutil

from ada.fem.formats.utils import LocalExecute


def run_abaqus(
    inp_path,
    cpus=2,
    gpus=None,
    run_ext=False,
    metadata=None,
    subr_path=None,
    execute=True,
    return_bat_str=False,
    exit_on_complete=True,
    run_in_shell=False,
):
    gpus = "" if gpus is None else f"GPUS={gpus}"
    run_cmd = None
    custom_bat_str = None
    if subr_path is not None:
        run_cmd, custom_bat_str = create_subroutine_input(inp_path, subr_path, cpus, gpus)

    aba_exe = AbaqusExecute(
        inp_path, cpus=cpus, run_ext=run_ext, metadata=metadata, auto_execute=execute, run_in_shell=run_in_shell
    )
    return aba_exe.run(exit_on_complete, run_cmd=run_cmd, bat_start_str=custom_bat_str)


class AbaqusExecute(LocalExecute):
    def run(self, exit_on_complete=True, run_cmd=None, bat_start_str=None):
        from ada.fem.formats.general import FEATypes

        exe_path = self.get_exe(FEATypes.ABAQUS)
        gpus = "" if self._gpus is None else f"GPUS={self._gpus}"
        if run_cmd is None:
            run_cmd = f"{exe_path} job={self.analysis_name} CPUS={self._cpus}{gpus} interactive"
        stop_cmd = f"abaqus terminate job={self.analysis_name}"
        out = self._run_local(run_cmd, stop_cmd, exit_on_complete, bat_start_str)
        return out


def create_subroutine_input(inp_path, subroutine_path, cpus, gpus):
    inp_path = pathlib.Path(inp_path)
    if inp_path.exists() is False:
        raise FileNotFoundError(f'Unable to find inp file "{inp_path}"')
    analysis_name = inp_path.name.replace(".inp", "")
    subroutine_path = pathlib.Path(subroutine_path)
    subr_name = subroutine_path.with_suffix("").name
    shutil.copy(subroutine_path, inp_path.parent / subroutine_path.name)
    subr = f"user={subr_name}"
    param = ["job=" + analysis_name, subr, "CPUS=" + str(cpus), gpus, "interactive"]
    run_cmd = " ".join([str(val) for val in param if str(val) != ""])
    prog = r"C:\Program Files (x86)"
    custom_bat_str = rf'''
call "{prog}\IntelSWTools\compilers_and_libraries_2018.3.210\windows\bin\ipsxe-comp-vars.bat" intel64 vs2017
call "{prog}\Microsoft Visual Studio 11.0\VC\bin\amd64\vcvars64.bat" intel 64
call "C:\SIMULIA\CAE\2017\win_b64\code\bin\ABQLauncher.exe"'''
    return run_cmd, custom_bat_str



import os
import pathlib
import shutil

import plotly.graph_objs as go
from plotly import io as pio


def easy_plotly(
    title,
    in_data,
    xlbl="X-axis",
    ylbl="Y-axis",
    xrange=None,
    yrange=None,
    yaxformat="E",
    legend=None,
    autoreverse=False,
    save_filename=None,
    mode="lines",
    marker="circle",
    traces=None,
    template="plotly_white",
    annotations=None,
    shapes=None,
    renderer="notebook_connected",
    return_widget=True,
):
    """
    A Plotly template for quick and easy interactive scatter plotting using some pre-defined values. If you need more
    control of the plotly plot, you are probably better off using plotly directly

    See https://plot.ly/python/reference/#scatter for a complete list of input for

    :param title: Plot title
    :param in_data: tuple (x, y) for single plots or dict {'var1':{'x': [..], 'y': [..] }, 'var2': {..}, etc..}
    :param xlbl: X-axis label
    :param ylbl: Y-axis label
    :param xrange: min and max values of x-axis
    :param yrange: min and max values of y-axis
    :param yaxformat: "none" | "e" | "E" | "power" | "SI" | "B" (default) exponent format of y-axis.
    :param legend: dict(x=-.1, y=1.2)
    :param autoreverse: Autoreverse the X-axis (opposed to inputting the reversed x-list)
    :param save_filename: Abs path to file location or file name of figure.
    :param mode:
    :param marker:
    :param traces: Add plotly traces manually
    :param template: Which plot template. Default is 'plotly_white'. Alternatives are shown below
    :param annotations:
    :param renderer: Which renderer should be used. Default is 'notebook_connected'. See below for alternatives
    :param return_widget:
    :type title: str
    :type xlbl: str
    :type ylbl: str
    :type xrange: list
    :type yrange: list
    :type yaxformat: str
    :type save_filename: str
    :type mode: str

    Templates:
                'ggplot2', 'seaborn', 'plotly', 'plotly_white', 'plotly_dark', 'presentation', 'xgridoff', 'none'

    renderers:
                'plotly_mimetype', 'jupyterlab', 'nteract', 'vscode', 'notebook', 'notebook_connected', 'kaggle',
                'azure', 'colab', 'cocalc', 'databricks', 'json', 'png', 'jpeg', 'jpg', 'svg', 'pdf', 'browser',
                'firefox', 'chrome', 'chromium', 'iframe', 'iframe_connected', 'sphinx_gallery'

    """

    plot_data = []
    if type(in_data) is dict:
        for key in in_data.keys():
            if type(in_data[key]) is dict:
                x_ = in_data[key]["x"]
                y_ = in_data[key]["y"]
            elif type(in_data[key]) is tuple:
                x_ = in_data[key][0]
                y_ = in_data[key][1]
            else:
                raise Exception('unrecognized input in dict "{}"'.format(type(in_data[key])))

            trace = go.Scatter(
                x=x_,
                y=y_,
                name=key,
                mode=mode,
                marker=dict(symbol=marker),
            )
            plot_data.append(trace)
    elif type(in_data) in [list, tuple]:
        x, y = in_data
        trace = go.Scatter(
            x=x,
            y=y,
            mode=mode,
            marker=dict(symbol=marker),
        )
        plot_data.append(trace)
    else:
        if traces is None:
            raise Exception('No Recognized input type found for "in_data" or "traces"')
    if traces is not None:
        plot_data += traces
    autorange = "reversed" if autoreverse is True else None
    layout = go.Layout(
        title=title,
        xaxis=dict(
            title=xlbl,
            titlefont=dict(family="Arial, monospace", size=18, color="#7f7f7f"),
            autorange=autorange,
            range=xrange,
        ),
        yaxis=dict(
            title=ylbl,
            titlefont=dict(family="Arial, monospace", size=18, color="#7f7f7f"),
            range=yrange,
            exponentformat=yaxformat,
        ),
        legend=legend,
        template=template,
        shapes=shapes,
    )
    if annotations is not None:
        layout["annotations"] = annotations
    fig = go.FigureWidget(data=plot_data, layout=layout)
    # plotly.offline.init_notebook_mode(connected=True)
    if save_filename is not None:
        # fig.show(renderer=renderer)
        filepath = save_filename
        if ".png" not in filepath:
            filepath += ".png"

        dirpath = os.path.dirname(filepath)
        print('Saving "{}" to "{}"'.format(os.path.basename(filepath), dirpath))
        filename = os.path.splitext(filepath)[0].replace(dirpath + "\\", "")
        if os.path.isdir(dirpath) is False:
            os.makedirs(dirpath)
        pio.write_image(fig, save_filename, width=1600, height=800)
        if "\\" not in save_filename:
            output_file = pathlib.Path(f"C:/ADA/temp/{filename}.png")
            if os.path.isfile(output_file) is True:
                shutil.move(output_file, dirpath + "\\" + filename + ".png")
            else:
                print("{} not found".format(output_file))

    else:
        if return_widget is True:
            return fig
        fig.show(renderer=renderer)



import logging
from typing import Union

import numpy as np

from .categories import SectionCat
from .concept import GeneralProperties, Section

# List of documents the various formulas are based upon
#
#   * StructX.com (https://www.structx.com/geometric_properties.html)
#   * DNVGL. (2011). Appendix B Section properties & consistent units Table of Contentsec. I.
#   * W. Beitz, K.H. Küttner: "Dubbel, Taschenbuch für den Maschinenbau" 17. Auflage (17th ed.)
#     Springer-Verlag 1990
#   * Arne Selberg: "Stålkonstruksjoner" Tapir 1972
#   * sec. Timoshenko: "Strength of Materials, Part I, Elementary Theory and Problems" Third Edition 1995 D.
#     Van Nostrand Company Inc.


def calculate_general_properties(section: Section) -> Union[None, GeneralProperties]:
    """Calculations of cross section properties are based on different sources of information."""
    bt = SectionCat.BASETYPES
    section_map = {
        bt.CIRCULAR: calc_circular,
        bt.IPROFILE: calc_isec,
        bt.BOX: calc_box,
        bt.TUBULAR: calc_tubular,
        bt.ANGULAR: calc_angular,
        bt.CHANNEL: calc_channel,
        bt.FLATBAR: calc_flatbar,
        bt.TPROFILE: calc_isec,
    }

    base_type = SectionCat.get_shape_type(section)

    if base_type == bt.GENERAL:
        logging.error("Re-Calculating a general section")
        return None

    calc_func = section_map.get(base_type, None)

    if calc_func is None:
        raise Exception(
            f'Section type "{section.type}" is not yet supported in the cross section parameter calculations'
        )

    return calc_func(section)


def calc_box(sec: Section) -> GeneralProperties:
    """Calculate box cross section properties"""

    sfy = 1.0
    sfz = 1.0

    Ax = sec.w_btn * sec.t_fbtn + sec.w_top * sec.t_ftop + sec.t_w * (sec.h - (sec.t_fbtn + sec.t_ftop)) * 2

    by = sec.w_top
    tt = sec.t_ftop
    tb = sec.t_fbtn
    ty = sec.t_w
    hz = sec.h

    a = tb / 2
    b = (hz + tb - tt) / 2
    c = hz - tt / 2
    d = sec.h - sec.t_fbtn - sec.t_ftop
    e = by * tb
    f = by * tt
    g = ty * d

    area = e + f + 2 * g
    h = (e * a + f * c + 2 * b * g) / area
    ha = sec.h - (sec.t_fbtn + sec.t_ftop) / 2.0
    hb = sec.w_top - sec.t_w

    Ix = 4 * (ha * hb) ** 2 / (hb / tb + hb / ty + 2 * ha / ty)
    Iy = (by * (tb ** 3 + tt ** 3) + 2 * ty * d ** 3) / 12 + e * (h - a) ** 2 + f * (c - h) ** 2 + 2 * g * (b - h) ** 2

    Iz = ((sec.t_fbtn + sec.t_ftop) * sec.w_top ** 3 + 2 * d * sec.t_w ** 3) / 12 + (g * hb ** 2) / 2
    Iyz = 0
    Wxmin = Ix * (hb + ha) / (ha * hb)
    Wymin = Iy / max(sec.h - h, h)
    Wzmin = 2 * Iz / sec.w_top
    Sy = e * (h - a) + ty * (h - tb) ** 2
    Sz = (sec.t_fbtn + sec.t_ftop) * sec.w_top ** 2 / 8 + g * hb / 2
    Shary = (Iz / Sz) * 2 * sec.t_w * sfy
    Sharz = (Iy / Sy) * 2 * ty * sfz
    Shceny = 0
    Shcenz = c - h - sec.t_fbtn * ha / (sec.t_fbtn + sec.t_ftop)
    Cy = sec.w_top / 2
    Cz = h
    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=sfy,
        Sfz=sfz,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_isec(sec: Section) -> GeneralProperties:
    """Calculate I/H cross section properties"""

    sfy = 1.0
    sfz = 1.0
    hz = sec.h
    bt = sec.w_top
    tt = sec.t_ftop
    ty = sec.t_w
    bb = sec.w_btn
    tb = sec.t_fbtn

    Ax = bt * tt + ty * (hz - (tb + tt)) + bb * tb
    hw = hz - tt - tb
    a = tb + hw + tt / 2
    b = tb + hw / 2
    c = tb / 2

    z = (bt * tt * a + hw * ty * b + bb * tb * c) / Ax

    tra = (bt * tb ** 3) / 12 + bt * tt * (hz - tt / 2 - z) ** 2
    trb = (ty * hw ** 3) / 12 + ty * hw * (tb + hw / 2 - z) ** 2
    trc = (bb * tb ** 3) / 12 + bb * tb * (tb / 2 - z) ** 2

    if tt == ty and tt == tb:
        Ix = (tt ** 3) * (hw + bt + bb - 1.2 * tt) / 3
        Wxmin = Ix / tt
    else:
        Ix = 1.3 * (bt * tt ** 3 + hw * ty ** 3 + bb * tb ** 3) / 3
        Wxmin = Ix / max(tt, ty, tb)

    Iy = tra + trb + trc
    Iz = (tb * bb ** 3 + hw * ty ** 3 + tt * bt ** 3) / 12
    Iyz = 0
    Wymin = Iy / max(hz - z, z)
    Wzmin = 2 * Iz / max(bb, bt)

    # Sy should be checked. Confer older method implementation.
    # Sy = sum(x_i * A_i)
    # Sy = (((tt * bt) ** 2) * (hw / 2 + tt / 2)) * 2
    Sy = Iy / (sec.w_top / 2)

    # Sy = (sec.t_w*sec.h/2)(sec.h/2)
    Sz = (tt * bt ** 2 + tb * bb ** 2 + hw * ty ** 2) / 8
    Shary = (Iz / Sz) * (tb + tt) * sfy
    Sharz = (Iy / Sy) * ty * sfz
    Shceny = 0
    Shcenz = ((hz - tt / 2) * tt * bt ** 3 + (tb ** 2) * (bb ** 3) / 2) / (tt * bt ** 3 + tb * bb ** 3) - z
    Cy = bb / 2
    Cz = z

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=1,
        Sfz=1,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_angular(sec: Section) -> GeneralProperties:
    """Calculate L cross section properties"""

    # rectangle A properties (web)
    a_w = sec.t_w
    a_h = sec.h - sec.t_fbtn
    a_dy = a_w / 2
    a_dz = sec.t_fbtn + a_h / 2
    a_area = a_w * a_h

    # rectangle B properties (flange)
    b_w = sec.w_btn
    b_h = sec.t_fbtn
    b_dy = b_w / 2
    b_dz = b_h / 2
    b_area = b_h * b_w

    # Find centroid
    c_y = (a_area * a_dy + b_area * b_dy) / (a_area + b_area)
    c_z = (a_area * a_dz + b_area * b_dz) / (a_area + b_area)

    # c_z_opp = sec.h - c_z

    a_dcy = a_dy - c_y
    b_dcy = b_dy - c_y

    a_dcz = a_dz - c_z
    b_dcz = b_dz - c_z

    # Iz_a + A_a*dcy_a**2

    Iz_a = (1 / 12) * a_h * a_w ** 3 + a_area * a_dcy ** 2
    Iz_b = (1 / 12) * b_h * b_w ** 3 + b_area * b_dcy ** 2
    Iz = Iz_a + Iz_b

    Iy_a = (1 / 12) * a_w * a_h ** 3 + a_area * a_dcz ** 2
    Iy_b = (1 / 12) * b_w * b_h ** 3 + b_area * b_dcz ** 2
    Iy = Iy_a + Iy_b

    posweb = False

    r = 0

    hz = sec.h
    ty = sec.t_w
    tz = sec.t_fbtn
    by = sec.w_btn

    sfy = 1.0
    sfz = 1.0
    hw = hz - tz
    b = tz - hw / 2.0
    c = tz / 2.0
    piqrt = np.arctan(1.0)
    Ax = ty * hw + by * tz + (1 - piqrt) * r ** 2
    y = (hw * ty ** 2 + tz * by ** 2) / (2 * Ax)
    z = (hw * b * ty + tz * by * c) / Ax
    d = 6 * r + 2 * (ty + tz - np.sqrt(4 * r * (2 * r + ty + tz) + 2 * ty * tz))
    e = hw + tz - z
    f = hw - e
    ri = y - ty
    rj = by - y
    rk = ri + 0.5 * ty
    rl = z - c

    if tz >= ty:
        h = hw
    else:
        raise ValueError("Currently not implemented this yet")

    Ix = (1 / 3) * (by * tz ** 3 + (hz - tz) * ty ** 3)
    Iyz = (rl * tz / 2) * (y ** 2 - rj ** 2) - (rk * ty / 2) * (e ** 2 - f ** 2)

    Wxmin = Ix / d
    Wymin = Iy / max(z, hz - h)
    Wzmin = Iz / max(y, rj)
    Sy = (ty * e ** 2) / 2
    Sz = (tz * rj ** 2) / 2
    Shary = (Iz * tz / Sz) * sfy
    Sharz = (Iy * tz / Sy) * sfz

    if posweb:
        Iyz = -Iyz
        Shceny = rk
        Cy = by - y
    else:
        Shceny = -rk
        Cy = y
    Shcenz = -rl
    Cz = z

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=1,
        Sfz=1,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_tubular(sec: Section) -> GeneralProperties:
    """Calculate Tubular cross section properties"""

    t = sec.wt
    sfy = 1.0
    sfz = 1.0

    dy = sec.r * 2
    di = dy - 2 * t
    Ax = np.pi * sec.r ** 2 - np.pi * (sec.r - t) ** 2
    Ix = 0.5 * np.pi * ((dy / 2) ** 4 - (di / 2) ** 4)
    Iy = Ix / 2
    Iz = Iy
    Iyz = 0
    Wxmin = 2 * Ix / dy
    Wymin = 2 * Iy / dy
    Wzmin = 2 * Iz / dy
    Sy = (dy ** 3 - di ** 3) / 12
    Sz = Sy
    Shary = (2 * Iz * t / Sy) * sfy
    Sharz = (2 * Iy * t / Sz) * sfz
    Shceny = 0
    Shcenz = 0
    Cy = 0.0
    Cz = 0.0

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=1,
        Sfz=1,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_circular(sec: Section) -> GeneralProperties:
    Sfy = 1.0
    Sfz = 1.0
    Iyz = 0.0

    Ax = np.pi * sec.r ** 2
    Iy = (np.pi * sec.r ** 4) / 4
    Iz = Iy
    Ix = 0.5 * np.pi * sec.r ** 4
    Wymin = 0.25 * np.pi * sec.r ** 3
    Wzmin = Wymin

    Wxmin = Ix / sec.r

    t = sec.r * 0.99
    dy = sec.r * 2
    di = dy - 2 * t
    Sy = (dy ** 3 - di ** 3) / 12
    Sz = Sy
    Shary = (2 * Iz * t / Sy) * Sfy
    Sharz = (2 * Iy * t / Sz) * Sfz
    Shceny = 0
    Shcenz = 0
    Cy = 0.0
    Cz = 0.0

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=Sfy,
        Sfz=Sfz,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_flatbar(sec: Section) -> GeneralProperties:
    """Flatbar (not supporting unsymmetric profile)"""
    w = sec.w_btn
    hz = sec.h

    a = 0.0
    h = hz * w / (2 * w)
    b = w / 2
    d = 0

    Sfy = 1.0
    Sfz = 1.0

    Ax = w * hz
    Iy = w * hz ** 3 / 12
    Iz = hz * w ** 3 / 12

    bm = 2 * w * hz ** 2 / (hz ** 2 + Ax ** 2)
    Wymin = Iy / max(h, d)
    Wzmin = 2 * Iz / max(w, w)
    Iyz = 0.0
    if hz == bm:
        ca = 0.141
        cb = 0.208
        Ix = ca * hz ** 4
        Wxmin = cb * hz ** 3
    elif hz < bm:
        cn = bm / hz
        ca = (1 - 0.63 / cn + 0.052 / cn ** 5) * 3
        cb = ca / (1 - 0.63 / (1 + cn ** 3))
        Ix = ca * bm * hz ** 3
        Wxmin = cb * bm * hz ** 2
    else:
        cn = hz / bm
        ca = (1 - 0.63 / cn + 0.052 / cn ** 5) * 3
        cb = ca / (1 - 0.63 / (1 + cn ** 3))
        Ix = ca * hz * bm ** 3
        Wxmin = cb * hz * bm ** 3

    Sy = (w * h ** 2) / 2 + (b - w / 2) * (h ** 2) / 3
    Sz = hz * ((w ** 2) / 8 + a * (w / 4 + a / 6))

    Shary = Iz * hz * Sfy / Sz
    Sharz = 2 * Iy * b * Sfz / Sy

    Shceny = 0.0
    Shcenz = 0.0
    Cy = w / 2
    Cz = hz

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=Sfy,
        Sfz=Sfz,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )


def calc_channel(sec: Section) -> GeneralProperties:
    """Calculate section properties of a channel profile"""
    posweb = False
    hz = sec.h
    ty = sec.t_w
    tz = sec.t_fbtn
    by = sec.w_btn
    sfy = 1.0
    sfz = 1.0

    a = hz - 2 * tz
    Ax = 2 * by * tz + a * ty
    y = (2 * tz * by ** 2 + a * ty ** 2) / (2 * Ax)
    Iy = (ty * a ** 3) / 12 + 2 * ((by * tz ** 3) / 12 + by * tz * ((a + tz) / 2) ** 2)

    if tz == ty:
        Ix = ty ** 3 * (2 * by + a - 2.6 * ty) / 3
        Wxmin = Ix / Iy
    else:
        Ix = 1.12 * (2 * by * tz ** 3 + a * ty ** 3) / 3
        Wxmin = Ix / max(tz, ty)

    Iz = 2 * ((tz * by ** 3) / 12 + tz * by * (by / 2 - y) ** 2) + (a * ty ** 3) / 12 + a * ty * (y - ty / 2) ** 2
    Iyz = 0
    Wymin = 2 * Iy / hz
    Wzmin = Iz / max(by - y, y)
    Sy = by * tz * (tz + a) / 2 + (ty * a ** 2) / 8
    Sz = tz * (by - y) ** 2

    Shary = (Iz / Sz) * (2 * tz) * sfy
    Sharz = (Iy / Sy) * ty * sfz

    if tz == ty:
        q = ((by - ty / 2) ** 2) * ((hz - tz) ** 2) * tz / 4 * Iy
    else:
        q = ((by - ty / 2) ** 2) * tz / (2 * (by - ty / 2) * tz + (hz - tz) * ty / 3)

    if posweb:
        Shceny = y - ty / 2 + q
        Cy = by
    else:
        Shceny = -(y - ty / 2 + q)
        Cy = by - y

    Cz = hz / 2
    Shcenz = 0

    return GeneralProperties(
        Ax=Ax,
        Ix=Ix,
        Iy=Iy,
        Iz=Iz,
        Iyz=Iyz,
        Wxmin=Wxmin,
        Wymin=Wymin,
        Wzmin=Wzmin,
        Shary=Shary,
        Sharz=Sharz,
        Shceny=Shceny,
        Shcenz=Shcenz,
        Sy=Sy,
        Sz=Sz,
        Sfy=sfy,
        Sfz=sfz,
        Cy=Cy,
        Cz=Cz,
        parent=sec,
    )



from ada import FEM

from .write_utils import write_ff


def mass_str(fem: FEM) -> str:
    out_str = ""

    for mass in fem.masses.values():
        for m in mass.fem_set.members:
            if mass.type == mass.TYPES.MASS:
                if mass.point_mass_type == mass.PTYPES.ISOTROPIC:
                    masses = [mass.mass for _ in range(0, 3)] + [0, 0, 0]
                elif mass.point_mass_type == mass.PTYPES.ANISOTROPIC:
                    masses = mass.mass + [0, 0, 0]
                else:
                    raise NotImplementedError(f"Mass point mass type {mass.point_mass_type} is not yet supported")
            else:
                raise NotImplementedError(f"Mass type {mass.type} is not yet supported")
            data = (tuple([m.id, 6] + masses[:2]), tuple(masses[2:]))
            out_str += write_ff("BNMASS", data)
    return out_str


def write_node_with_mass_point():
    raise NotImplementedError()



from dataclasses import dataclass


class StabilizeTypes:
    ENERGY = "energy"
    DAMPING = "damping"
    CONTINUE = "continue"


@dataclass
class Stabilize:
    factor: float
    allsdtol: float
    stabilize_type: StabilizeTypes = StabilizeTypes.ENERGY
    energy: float = None
    damping: float = None
    stabilize_continue: bool = True

    def to_input_str(self):
        st = StabilizeTypes
        stable_map = {
            st.ENERGY: f"stabilize={self.factor}, allsdtol={self.allsdtol}",
            st.DAMPING: f"stabilize, factor={self.factor}, allsdtol={self.allsdtol}",
            st.CONTINUE: "stabilize, continue=ON",
        }
        stabilize_str = stable_map.get(self.stabilize_type, None)
        if stabilize_str is None:
            raise ValueError(f'Unrecognized stabilization type "{self.stabilize_type}"')

        return stabilize_str


@dataclass
class SolverOptionsAbaqus:

    init_accel_calc: bool = True
    restart_int: int = None
    unsymm: bool = False
    stabilize: Stabilize = None

    """
    :param init_accel_calc: Calculate Initial acceleration in the beginning of the step
    :param restart_int: Restart interval
    :param unsymm: Unsymmetric Matrix storage (default=False)
    :param stabilize: Default=None.
    """



from __future__ import annotations

import logging
import mmap
import os
import pathlib
import re
from dataclasses import dataclass
from itertools import chain
from typing import Iterable, List, Union

import numpy as np

from ada.concepts.containers import Nodes
from ada.concepts.levels import FEM, Assembly, Part
from ada.concepts.points import Node
from ada.concepts.structural import Material
from ada.concepts.transforms import Rotation, Transform
from ada.core.utils import Counter, roundoff
from ada.fem import (
    Bc,
    Connector,
    ConnectorSection,
    Constraint,
    Csys,
    Elem,
    FemSection,
    FemSet,
    Interaction,
    InteractionProperty,
    Mass,
    PredefinedField,
    Surface,
)
from ada.fem.containers import FemElements, FemSections, FemSets
from ada.fem.formats.abaqus.common import AbaCards
from ada.fem.interactions import ContactTypes
from ada.fem.shapes import ElemShape, ElemType
from ada.materials.metals import CarbonSteel

from ..utils import str_to_int
from . import cards

part_name_counter = Counter(1, "Part")
_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL


@dataclass
class InstanceData:
    part_ref: str
    instance_name: str
    instance_bulk: str
    transform: Transform = Transform()


def read_fem(assembly: Assembly, fem_file, fem_name=None):
    """This will create and add an AbaqusPart object based on a path reference to a Abaqus input file."""

    print("Starting import of Abaqus input file")

    if fem_name is not None:
        global part_name_counter
        part_name_counter = Counter(1, fem_name)

    bulk_str = read_bulk_w_includes(fem_file)
    lbulk = bulk_str.lower()
    ass_start = lbulk.find("\n*assembly")
    ass_end = lbulk.rfind("\n*end assembly")
    step_start = lbulk.rfind("\n*step")

    ass_start = ass_start + 1 if ass_start != -1 else ass_start
    ass_end = ass_end + 1 if ass_end != -1 else ass_end
    step_start = step_start + 1 if step_start != -1 else step_start

    if ass_start == -1 and ass_end == -1:
        uses_assembly_parts = False
        assembly_str = bulk_str
        props_str = bulk_str
    else:
        uses_assembly_parts = True
        assembly_str = bulk_str[ass_start : ass_end + 2]
        props_str = bulk_str[ass_end + 2 : step_start]

    inst_end = assembly_str.lower().rfind("\n*end instance")
    inst_end = inst_end + 15 if inst_end != -1 else inst_end

    get_materials_from_bulk(assembly, props_str)
    get_intprop_from_lines(assembly, props_str)

    ass_data = extract_instance_data(assembly_str[:inst_end])

    part_list = import_parts(bulk_str[:ass_start], ass_data, assembly)
    if len(part_list) == 0:
        add_fem_without_assembly(bulk_str, assembly)

    if uses_assembly_parts is True:
        ass_sets = assembly_str[inst_end:]
        assembly.fem.nodes += get_nodes_from_inp(ass_sets, assembly.fem)
        assembly.fem.lcsys.update(get_lcsys_from_bulk(ass_sets, assembly.fem))
        assembly.fem.elements += get_elem_from_inp(ass_sets, assembly.fem)
        assembly.fem.sets += get_sets_from_bulk(ass_sets, assembly.fem)
        assembly.fem.sets.link_data()
        assembly.fem.surfaces.update(get_surfaces_from_bulk(ass_sets, assembly.fem))
        assembly.fem.constraints += get_constraints_from_inp(ass_sets, assembly.fem)
        assembly.fem.connector_sections.update(get_connector_sections_from_bulk(props_str, assembly.fem))
        assembly.fem.connectors.update(get_connectors_from_inp(ass_sets, assembly.fem))
        assembly.fem.bcs += get_bcs_from_bulk(props_str, assembly.fem)

    add_interactions_from_bulk_str(props_str, assembly)
    get_initial_conditions_from_lines(assembly, props_str)


def read_bulk_w_includes(inp_path) -> str:
    re_bulk_include = re.compile(r"\*Include,\s*input=(.*?)$", _re_in)
    bulk_repl = dict()
    with open(inp_path, "r") as inpDeck:
        bulk_str = inpDeck.read()
        for m in re_bulk_include.finditer(bulk_str):
            search_key = m.group(0)
            with open(os.path.join(os.path.dirname(inp_path), m.group(1)), "r") as d:
                bulk_repl[search_key] = d.read()

    for key, val in bulk_repl.items():
        bulk_str = bulk_str.replace(key, val)
    return bulk_str


def import_bulk(file_path, buffer_function):
    with open(file_path, "r") as f:
        return buffer_function(f.read())


def import_bulk2(file_path, buffer_function):
    with open(file_path, "r") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as m:
            return buffer_function(m)


def extract_instance_data(assembly_bulk) -> dict[str, List[InstanceData]]:
    ass_data = {}
    for m in cards.inst_matches.finditer(assembly_bulk):
        d = m.groupdict()
        inst_name, part_name, bulk_str = d["inst_name"], d["part_name"], d["bulk_str"]
        inst_data = get_instance_data(inst_name, part_name, bulk_str)
        if inst_data.part_ref not in ass_data.keys():
            ass_data[inst_data.part_ref] = []
        ass_data[inst_data.part_ref].append(inst_data)

    return ass_data


def import_parts(bulk_str, instance_data: dict[str, List[InstanceData]], assembly: Assembly) -> List[Part]:
    part_list = []

    for m in cards.parts_matches.finditer(bulk_str):
        d = m.groupdict()
        name = d.get("name")
        part_bulk_str = d.get("bulk_str")

        for i in instance_data[name]:
            p_bulk_str = i.instance_bulk if part_bulk_str == "" and i.instance_bulk != "" else part_bulk_str
            part = get_fem_from_bulk_str(name, p_bulk_str, assembly, i)
            part_list.append(part)
    return part_list


def add_fem_without_assembly(bulk_str, assembly: Assembly) -> Part:
    part_name_matches = list(cards.part_names.finditer(bulk_str))
    p_nmatch = tuple(part_name_matches)

    if len(p_nmatch) != 1:
        p_bulk = bulk_str
        p_name = None
    else:
        p_name = p_nmatch[0].group(1)
        p_bulk = p_nmatch[0].group(2)

    p_name = next(part_name_counter) if p_name is None else p_name
    inst = InstanceData("", p_name, "")

    return get_fem_from_bulk_str(p_name, p_bulk, assembly, inst)


def get_fem_from_bulk_str(name, bulk_str, assembly: Assembly, instance_data: InstanceData) -> Part:
    instance_name = name if instance_data.instance_name is None else instance_data.instance_name
    part = assembly.add_part(Part(name, fem=FEM(name=instance_name)))
    fem = part.fem
    fem.nodes = get_nodes_from_inp(bulk_str, fem)
    fem.nodes.move(move=instance_data.transform.translation, rotate=instance_data.transform.rotation)
    fem.elements = get_elem_from_inp(bulk_str, fem)
    fem.elements.build_sets()
    fem.sets += get_sets_from_bulk(bulk_str, fem)
    fem.sections = get_sections_from_inp(bulk_str, fem)
    fem.bcs += get_bcs_from_bulk(bulk_str, fem)
    fem.masses = get_mass_from_bulk(bulk_str, fem)
    fem.surfaces.update(get_surfaces_from_bulk(bulk_str, fem))
    fem.lcsys = get_lcsys_from_bulk(bulk_str, fem)
    fem.constraints = get_constraints_from_inp(bulk_str, fem)

    print(8 * "-" + f'Imported "{part.fem.instance_name}"')

    return part


def get_initial_conditions_from_lines(assembly: Assembly, bulk_str: str):
    """
    TODO: Optimize this function


    ** PREDEFINED FIELDS
    **
    ** Name: IC-1   Type: Velocity
    *Initial Conditions, type=VELOCITY
    CONTAINER20FT-10000KG, 1, 0.
    CONTAINER20FT-10000KG, 2, 0.
    CONTAINER20FT-10000KG, 3, 4.1
    """
    if bulk_str.find("*Initial Conditions") == -1:
        return
    re_str = r"(?:\*\*\s*Name:\s*(.*?)\s*Type:\s*(.*?)\n)+\*Initial Conditions, type=.*?\n(?<=)((?:.*?)(?=\*|\Z))"

    def sort_props(line):
        ev = [x.strip() for x in line.split(",")]
        set_name = ev[0]
        try:
            dofs = int(ev[1])
        except BaseException as e:
            logging.debug(e)
            dofs = ev[1]
        if len(ev) > 2:
            magn = ev[2]
        else:
            magn = None
        return set_name, dofs, magn

    def grab_init_props(m):
        bc_name = m.group(1)
        bc_type = m.group(2)
        bcs = m.group(3)
        props = [sort_props(line) for line in bcs.splitlines()]
        set_name, dofs, magn = list(zip(*props))
        fem_set = None
        set_name_up = set_name[0]
        if set_name_up in assembly.fem.elsets.keys():
            fem_set = assembly.fem.elsets[set_name_up]
        elif set_name_up in assembly.fem.nsets.keys():
            fem_set = assembly.fem.nsets[set_name_up]
        else:
            for p in assembly.get_all_parts_in_assembly():
                if set_name_up in p.fem.elsets.keys():
                    fem_set = p.fem.elsets[set_name_up]
        if fem_set is None:
            raise ValueError(f'Unable to find fem set "{set_name[0]}"')

        return PredefinedField(bc_name, bc_type, fem_set, dofs, magn, parent=assembly.fem)

    for match in re.finditer(re_str, bulk_str, _re_in):
        assembly.fem.add_predefined_field(grab_init_props(match))


def get_materials_from_bulk(assembly: Assembly, bulk_str):
    re_str = (
        r"(\*Material,\s*name=.*?)(?=\*|\Z)(?!\*Elastic|\*Density|\*Plastic|"
        r"\*Damage Initiation|\*Damage Evolution|\*Expansion)"
    )
    re_materials = re.compile(re_str, _re_in)
    for m in re_materials.finditer(bulk_str):
        mat = mat_str_to_mat_obj(m.group())
        assembly.add_material(mat)


def get_intprop_from_lines(assembly: Assembly, bulk_str):
    """
    *Surface Interaction, name=contactProp
    *Friction
    0.,
    *Surface Behavior, pressure-overclosure=HARD
    """
    re_str = r"(\*Surface Interaction, name=.*?)(?=\*)(?!\*Friction|\*Surface Behavior|\*Surface Smoothing)"
    # surf_interact = AbaFF("Surface Interaction", [("name=",), ("bulk>",)], [("Friction", [(), ("bulkFRIC>",)])])

    assembly.fem.metadata["surf_smoothing"] = []
    for m in AbaCards.surface_smoothing.regex.finditer(bulk_str):
        d = m.groupdict()
        assembly.fem.metadata["surf_smoothing"].append(d)

    for m in re.findall(re_str, bulk_str, _re_in):
        name = re.search(r"name=(.*?)\n", m, _re_in).group(1)
        re_fric = re.search(r"\*Friction\n(.*?),", m, _re_in)
        fric = None
        if re_fric is not None:
            fric = re_fric.group(1)
        props = dict(name=name, friction=fric)
        res = re.search(
            r"\*Surface Behavior,\s*(?P<btype>.*?)(?:=(?P<behaviour>.*?)(?:$\n(?P<tabular>.*?)(?:\*|\Z)|$)|$)",
            m,
            _re_in,
        )
        if res is not None:
            res = res.groupdict()
            btype = res["btype"]
            behave = res["behaviour"] if btype.upper() != "PENALTY" else btype
            tabular = res["tabular"]
            if tabular is not None:
                tab_list = []
                for line in tabular.splitlines():
                    tab_list.append(tuple(np.fromstring(line, dtype=float, sep=",")))
                tabular = tab_list
                props.update(dict(pressure_overclosure=behave, tabular=tabular))
        assembly.fem.add_interaction_property(InteractionProperty(**props))


def get_instance_data(inst_name, p_ref, inst_bulk) -> InstanceData:
    """Move/rotate data lines are specified here:

    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-instance.htm
    """

    move_rot = re.compile(r"(?:^\s*(.*?),\s*(.*?),\s*(.*?)$)", _re_in)
    transform: Union[Transform, None] = Transform()
    mr = move_rot.finditer(inst_bulk)
    if mr is not None:
        for j, mo in enumerate(mr):
            content = mo.group(0)
            if "*" in content or j == 2 or content == "":
                break
            if j == 0:
                transform.move = (float(mo.group(1)), float(mo.group(2)), float(mo.group(3)))
            if j == 1:
                r = [float(x) for x in mo.group(3).split(",")]
                origin = (float(mo.group(1)), float(mo.group(2)), r[0])
                vector = (r[1], r[2], r[3])
                angle = r[4]
                transform.rotate = Rotation(origin, vector, angle)

    return InstanceData(p_ref, inst_name, inst_bulk, transform)


def mat_str_to_mat_obj(mat_str) -> Material:
    rd = roundoff

    # Name
    name = re.search(r"name=(.*?)\n", mat_str, _re_in).group(1).split("=")[-1].strip()

    # Density
    density_ = re.search(r"\*Density\n(.*?)(?:,|$)", mat_str, _re_in)
    if density_ is not None:
        density = rd(density_.group(1).strip().split(",")[0].strip(), 10)
    else:
        print('No density flag found for material "{}"'.format(name))
        density = None

    # Elastic
    re_elastic_ = re.search(r"\*Elastic(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_elastic_ is not None:
        re_elastic = re_elastic_.group(2).strip().split(",")
        young, poisson = rd(re_elastic[0]), rd(re_elastic[1])
    else:
        print('No Elastic properties found for material "{name}"'.format(name=name))
        young, poisson = None, None

    # Plastic
    re_plastic_ = re.search(r"\*Plastic\n(.*?)(?:\*|\Z)", mat_str, _re_in)
    if re_plastic_ is not None:
        re_plastic = [tuple(x.split(",")) for x in re_plastic_.group(1).strip().splitlines()]
        sig_p = [rd(x[0]) for x in re_plastic]
        eps_p = [rd(x[1]) for x in re_plastic]
    else:
        eps_p, sig_p = None, None

    # Expansion
    re_zeta = re.search(r"\*Expansion(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_zeta is not None:
        zeta = float(re_zeta.group(2).split(",")[0].strip())
    else:
        zeta = 0.0

    # Return material object
    model = CarbonSteel(
        rho=density,
        E=young,
        v=poisson,
        eps_p=eps_p,
        zeta=zeta,
        sig_p=sig_p,
        plasticity_model=None,
    )
    return Material(name=name, mat_model=model)


def import_multiple_inps(input_files_dir):
    """
    Import a set of inp files from a folder

    :param input_files_dir:
    """

    def read_inp(fname):
        if ".inp" not in fname:
            return None
        else:
            with open(pathlib.Path(input_files_dir) / fname, "r") as d:
                return d.read()

    return "".join([x for x in map(read_inp, os.listdir(input_files_dir)) if x is not None])


def get_nodes_from_inp(bulk_str, parent: FEM) -> Nodes:
    """Extract node information from abaqus input file string"""
    re_no = re.compile(
        r"^\*Node\s*(?:,\s*nset=(?P<nset>.*?)\n|\n)(?P<members>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def getnodes(m):
        d = m.groupdict()
        res = np.fromstring(list_cleanup(d["members"]), sep=",", dtype=np.float64)
        res_ = res.reshape(int(res.size / 4), 4)
        members = [Node(n[1:4], int(n[0]), parent=parent) for n in res_]
        if d["nset"] is not None:
            parent.sets.add(FemSet(d["nset"], members, "nset", parent=parent))
        return members

    nodes = list(chain.from_iterable(map(getnodes, re_no.finditer(bulk_str))))

    return Nodes(nodes, parent=parent)


def get_elem_from_inp(bulk_str, fem: FEM) -> FemElements:
    re_el = re.compile(
        r"^\*Element,\s*type=(?P<eltype>.*?)(?:\n|,\s*elset=(?P<elset>.*?)\s*\n)(?<=)(?P<members>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def grab_elements(match):
        d = match.groupdict()
        eltype = d["eltype"]
        elset = d["elset"]
        members = d["members"]
        res = re.search("[a-zA-Z]", members)
        if eltype.upper() in ElemShape.TYPES.cube20 + ElemShape.TYPES.cube27 or res is None:
            if eltype.upper() in ElemShape.TYPES.cube20 + ElemShape.TYPES.cube27:
                temp = members.splitlines()
                ntext = "".join([l1.strip() + "    " + l2.strip() + "\n" for l1, l2 in zip(temp[:-1:2], temp[1::2])])
            else:
                ntext = d["members"]
            res = np.fromstring(ntext.replace("\n", ","), sep=",", dtype=int)
            n = ElemShape.num_nodes(eltype) + 1
            res_ = res.reshape(int(res.size / n), n)
            return [Elem(e[0], [fem.nodes.from_id(n) for n in e[1:]], eltype, elset, parent=fem) for e in res_]
        else:

            # TODO: This code needs to be re-worked!
            elems = []
            for li in members.splitlines():
                new_mem = []
                temp = li.split(",")
                elid = str_to_int(temp[0])
                for d in temp[1:]:
                    temp2 = [x.strip() for x in d.split(".")]
                    par_ = None
                    if len(temp2) == 2:
                        par, setr = temp2
                        pfems = []
                        parents = fem.parent.get_all_parts_in_assembly()
                        for p in parents:
                            assert isinstance(p, Part)
                            pfems.append(p.fem.name)
                            if p.fem.name == par:
                                par_ = p
                                break
                        if par_ is None:
                            raise ValueError(f'Unable to find parent for "{par}"')
                        r = par_.fem.nodes.from_id(str_to_int(setr))
                        if type(r) != Node:
                            raise ValueError("Node ID not found")
                        new_mem.append(r)
                    else:
                        r = fem.nodes.from_id(str_to_int(d))
                        if type(r) != Node:
                            raise ValueError("Node ID not found")
                        new_mem.append(r)
                elems.append(Elem(elid, new_mem, eltype, elset, parent=fem))
            return elems

    return FemElements(
        chain.from_iterable(map(grab_elements, re_el.finditer(bulk_str))),
        fem_obj=fem,
    )


def get_sections_from_inp(bulk_str, fem: FEM) -> FemSections:
    iter_beams = get_beam_sections_from_inp(bulk_str, fem)
    iter_shell = get_shell_sections_from_inp(bulk_str, fem)
    iter_solid = get_solid_sections_from_inp(bulk_str, fem)

    return FemSections(chain.from_iterable([iter_beams, iter_shell, iter_solid]), fem)


def str_to_ints(instr):
    try:
        return [int(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]
    except ValueError:
        return [str(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]


def get_sets_from_bulk(bulk_str, fem: FEM) -> FemSets:
    if fem.parent is not None:
        all_parts = fem.parent.get_all_parts_in_assembly()
    else:
        all_parts = []

    def get_parent_instance(instance) -> FEM:
        if instance is None or fem.parent is None:
            return fem
        elif instance is not None and type(fem.parent) is Assembly:
            for p in filter(lambda x: x.fem.instance_name == instance, all_parts):
                return p.fem
        else:
            raise ValueError(f'Unable to find instance "{instance}" amongst assembly parts')

    def get_set(match):
        name = match.group(2)
        set_type = match.group(1)
        internal = True if match.group(3) is not None else False
        instance = match.group(4)
        generate = True if match.group(5) is not None else False
        members_str = match.group(6)
        gen_mem = str_to_ints(members_str) if generate is True else []
        members = [] if generate is True else str_to_ints(members_str)
        metadata = dict(instance=instance, internal=internal, generate=generate, gen_mem=gen_mem)

        parent_instance = get_parent_instance(instance)

        fem_set = FemSet(
            name,
            members,
            set_type=set_type,
            metadata=metadata,
            parent=parent_instance,
        )
        return fem_set

    return FemSets([get_set(x) for x in cards.re_sets.finditer(bulk_str)], fem)

    # import concurrent.futures
    # with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
    #     futures = []
    #     for number in re_sets.finditer(bulk_str):
    #         future = executor.submit(get_set, number)
    #         futures.append(future)
    #     sets = [future.result() for future in concurrent.futures.as_completed(futures)]
    #
    # return FemSetsCollection(sets, parent)


def get_bcs_from_bulk(bulk_str, fem: FEM) -> List[Bc]:
    bc_counter = Counter(1, "bc")

    def get_dofs(content: str):
        set_name = None
        dofs = []
        magn = []
        if content.count("\n") == 1:
            temp = content.split(",")
            set_name = temp[0].strip()
            dof_in = temp[1].strip().replace("\n", "")
            if dof_in.lower() == "encastre":
                dofs = dof_in
            else:
                dof = str_to_int(temp[1])
                if len(temp) == 2:
                    dofs = [x if dof == x else None for x in range(1, 7)]
                else:
                    dof_end = str_to_int(temp[2])
                    dofs = [x if dof <= x <= dof_end else None for x in range(1, 7)]

        else:
            for line in content.splitlines():
                ev = [x.strip() for x in line.split(",")]
                set_name = ev[0]
                try:
                    dofs.append(int(ev[1]))
                except BaseException as e:
                    logging.debug(e)
                    dofs.append(ev[1])
                if len(ev) > 3:
                    magn.append(ev[2])
        magn = None if len(magn) == 0 else magn
        return set_name, dofs, magn

    def get_nset(part_instance_name, set_name):
        for p in fem.parent.parts.values():
            if p.fem.instance_name == part_instance_name:
                return p.fem.sets.get_nset_from_name(set_name)

    def get_bc(match):
        d = match.groupdict()
        bc_name = d["name"] if d["name"] is not None else next(bc_counter)
        bc_type = d["type"]
        set_name, dofs, magn = get_dofs(d["content"])

        if "." in set_name:
            part_instance_name, set_name = set_name.split(".")
            fem_set = get_nset(part_instance_name, set_name)
        else:
            if set_name in fem.nsets.keys():
                fem_set = fem.sets.get_nset_from_name(set_name)
            else:
                val = str_to_int(set_name)
                if val in fem.nodes.dmap.keys():
                    node = fem.nodes.from_id(val)
                    fem_set = FemSet(bc_name + "_set", [node], "nset", parent=fem)
                else:
                    raise ValueError(f'Unable to find set "{set_name}" in part {fem}')

        if fem_set is None:
            raise Exception("Unable to Find node set")

        props = dict()
        if bc_type is not None:
            props["bc_type"] = bc_type
        if magn is not None:
            props["magnitudes"] = magn

        return Bc(bc_name, fem_set, dofs, parent=fem, **props)

    return [get_bc(match_in) for match_in in cards.re_bcs.finditer(bulk_str)]


def get_mass_from_bulk(bulk_str, parent: FEM):
    """

    *MASS,ELSET=MASS3001
    2.00000000E+03,

    :return:
    """

    re_masses = re.compile(
        r"\*(?P<mass_type>Nonstructural Mass|Mass|Rotary Inertia),\s*elset=(?P<elset>.*?)"
        r"(?:,\s*type=(?P<ptype>.*?)\s*|\s*)(?:, units=(?P<units>.*?)|\s*)\n\s*(?P<mass>.*?)$",
        _re_in,
    )

    def map_element(el_set: FemSet, mass_prop):
        elem = el_set.members[0]
        elem.mass_prop = mass_prop

    def get_mass(match):
        d = match.groupdict()
        elset = parent.sets.get_elset_from_name(d["elset"])
        mass_type = d["mass_type"]
        p_type = d["ptype"]
        mass_ints = [str_to_int(x.strip()) for x in d["mass"].split(",") if x.strip() != ""]
        if len(mass_ints) == 1:
            mass_ints = mass_ints[0]
        units = d["units"]
        mass = Mass(d["elset"], elset, mass_ints, mass_type, p_type, units, parent=parent)
        map_element(elset, mass)
        return mass

    return {m.name: m for m in map(get_mass, re_masses.finditer(bulk_str))}


def get_surfaces_from_bulk(bulk_str, parent):
    """

    *Surface, type=ELEMENT, name=btn_surf
    _btn_surf_SPOS, SPOS

    :return:
    """
    from ada.fem.surfaces import SurfTypes

    def interpret_member(mem):
        msplit = mem.split(",")
        try:
            ref = str_to_int(msplit[0])
        except BaseException as e:
            logging.debug(e)
            ref = msplit[0].strip()

        return tuple([ref, msplit[1].strip()])

    surf_d = dict()

    for m in AbaCards.surface.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].strip()
        surf_type = d["type"].upper() if d["type"] is not None else "ELEMENT"
        members_str: str = d["bulk"]
        if members_str.count("\n") >= 1:
            id_refs = [interpret_member(m) for m in members_str.splitlines()]
            set_ref, set_id_ref = None, None
        else:
            id_refs = None
            res = [x.strip() for x in members_str.split(",")]
            if len(res) == 2:
                set_ref, set_id_ref = res
            else:
                set_ref = res[0]
                set_id_ref = 1.0

        if id_refs is None:
            if surf_type == SurfTypes.NODE:
                if set_id_ref == "":
                    fem_set = FemSet(f"n{set_ref}_set", [parent.nodes.from_id(int(set_ref))], "nset")
                    parent.add_set(fem_set)
                    weight_factor = 1.0
                else:
                    if "." in set_ref:
                        ssplit = set_ref.split(".")
                        parent_ = None
                        fem_set = None
                        for prt in parent.parent.get_all_parts_in_assembly():
                            if prt.fem.name == ssplit[0]:
                                parent_ = prt.fem
                                fem_set = prt.fem.nsets[ssplit[1]]
                                break
                        if parent_ is None:
                            raise ValueError(f'Unable to find parent FEM "{ssplit[0]}"')
                    else:
                        fem_set = parent.nsets[set_ref]
                    weight_factor = float(set_id_ref)
                el_face_index = None
            else:
                weight_factor = None
                el_face_index = int(set_id_ref.replace("S", "")) - 1
                fem_set = parent.sets.get_elset_from_name(set_ref)
        else:
            fem_set = None
            weight_factor = None
            el_face_index = None

        surf_d[name] = Surface(
            name,
            surf_type,
            fem_set,
            weight_factor,
            el_face_index,
            id_refs,
            parent=parent,
        )

    return surf_d


def get_lcsys_from_bulk(bulk_str: str, parent: FEM) -> dict[str, Csys]:
    """
    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-orientation.htm#simakey-r-orientation


    :param bulk_str:
    :param parent:
    :return:
    """
    lcsysd = dict()
    for m in AbaCards.orientation.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].replace('"', "")
        defi = d["definition"] if d["definition"] is not None else "COORDINATES"
        system = d["system"] if d["system"] is not None else "RECTANGULAR"
        if defi.upper() == "COORDINATES":
            coords = [
                (float(d["ax"]), float(d["ay"]), float(d["az"])),
                (float(d["bx"]), float(d["by"]), float(d["bz"])),
            ]
            if d["cx"] is not None:
                coords += [(float(d["cx"]), float(d["cy"]), float(d["cz"]))]
            lcsysd[name] = Csys(name, system=system, coords=coords, parent=parent)
        else:
            raise NotImplementedError(f'Orientation definition "{defi}" is not yet supported')

    return lcsysd


def get_constraints_from_inp(bulk_str: str, fem: FEM):
    """

    ** Constraint: Container_RigidBody
    *Rigid Body, ref node=container_rp, elset=container

    *MPC
     BEAM,    2007,     161
     BEAM,    2008,     162
    """

    # Rigid Bodies

    constraints = []
    rbnames = Counter(1, "rgb")
    conames = Counter(1, "co")

    for m in AbaCards.tie.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        msurf = grab_set_from_assembly(d["surf1"], fem, "surface")
        ssurf = grab_set_from_assembly(d["surf2"], fem, "surface")
        constraints.append(Constraint(name, Constraint.TYPES.TIE, msurf, ssurf, metadata=dict(adjust=d["adjust"])))

    for m in AbaCards.rigid_bodies.regex.finditer(bulk_str):
        d = m.groupdict()
        name = next(rbnames)
        ref_node = grab_set_from_assembly(d["ref_node"], fem, FemSet.TYPES.NSET)
        elset = grab_set_from_assembly(d["elset"], fem, FemSet.TYPES.ELSET)
        constraints.append(Constraint(name, Constraint.TYPES.RIGID_BODY, ref_node, elset, parent=fem))

    couplings = []
    for m in AbaCards.coupling.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        rn = d["ref_node"].strip()
        sf = d["surface"].strip()
        if rn.isnumeric():
            ref_set = FemSet(next(conames), [fem.nodes.from_id(int(rn))], FemSet.TYPES.NSET, parent=fem)
            fem.sets.add(ref_set)
        else:
            ref_set = fem.nsets[rn]

        surf = fem.surfaces[sf]

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=int)
        size = res.size
        cols = 2
        rows = int(size / cols)
        dofs = res.reshape(rows, cols)

        csys_name = d.get("orientation", None)
        if csys_name is not None:
            if csys_name not in fem.lcsys.keys():
                raise ValueError(f'Csys "{csys_name}" was not found on part {fem}')
            csys = fem.lcsys[csys_name]
        else:
            csys = None

        couplings.append(Constraint(name, Constraint.TYPES.COUPLING, ref_set, surf, csys=csys, dofs=dofs, parent=fem))

    # Shell to Solid Couplings
    sh2solids = []
    for m in AbaCards.sh2so_re.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        influence = d["influence_distance"]
        surf1 = grab_set_from_assembly(d["surf1"], fem, "surface")
        surf2 = grab_set_from_assembly(d["surf2"], fem, "surface")
        sh2solids.append(Constraint(name, Constraint.TYPES.SHELL2SOLID, surf1, surf2, influence_distance=influence))

    # MPC's
    mpc_dict = dict()
    mpc_re = re.compile(r"\*mpc\n(?P<bulk>.*?)^\*", _re_in)
    mpc_content = re.compile(r"\s*(?P<mpc_type>.*?),\s*(?P<master>.*?),\s*(?P<slave>.*?)$", _re_in)
    for match in mpc_re.finditer(bulk_str):
        d1 = match.groupdict()
        for subm in mpc_content.finditer(d1["bulk"]):
            d = subm.groupdict()
            mpc_type = d["mpc_type"]
            m = d["master"]
            s = d["slave"]
            if mpc_type not in mpc_dict.keys():
                mpc_dict[mpc_type] = []
            try:
                n1_ = str_to_int(m)
            except BaseException as e:
                logging.debug(e)
                n1_ = grab_set_from_assembly(m, fem, FemSet.TYPES.NSET)

            try:
                n2_ = str_to_int(s)
            except BaseException as e:
                logging.debug(e)
                n2_ = grab_set_from_assembly(s, fem, FemSet.TYPES.NSET)

            mpc_dict[mpc_type].append((n1_, n2_))

    def get_mpc(mpc_values):
        m_set, s_set = zip(*mpc_values)
        mpc_name = mpc_type + "_mpc"
        mset = FemSet("mpc_" + mpc_type + "_m", m_set, FemSet.TYPES.NSET)
        sset = FemSet("mpc_" + mpc_type + "_s", s_set, FemSet.TYPES.NSET)
        return Constraint(mpc_name, Constraint.TYPES.MPC, mset, sset, mpc_type=mpc_type, parent=fem)

    mpcs = [get_mpc(mpc_values_in) for mpc_values_in in mpc_dict.values()]

    return list(chain.from_iterable([constraints, couplings, sh2solids, mpcs]))


def get_connector_sections_from_bulk(bulk_str: str, parent: FEM) -> dict[str, ConnectorSection]:
    consecsd = dict()

    for m in AbaCards.connector_behaviour.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        comp = int(d["component"])

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=np.float64)
        size = res.size
        cols = comp + 1
        rows = int(size / cols)
        res_ = res.reshape(rows, cols)
        consecsd[name] = ConnectorSection(name, [res_], [], metadata=d, parent=parent)
    return consecsd


def get_connectors_from_inp(bulk_str, fem):
    """

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return:
    """

    nsuffix = Counter(1, "_")
    cons = dict()
    for m in AbaCards.connector_section.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["behavior"] + next(nsuffix)
        elset = fem.elsets[d["elset"]]
        elem = elset.members[0]

        csys_ref = d["csys"].replace('"', "")
        csys_ref = csys_ref[:-1] if csys_ref[-1] == "," else csys_ref

        con_type = d["contype"]
        if con_type[-1] == ",":
            con_type = con_type[:-1]

        n1 = elem.nodes[0]
        n2 = elem.nodes[1]

        cons[name] = Connector(
            name,
            elem.id,
            n1,
            n2,
            con_type,
            fem.connector_sections[d["behavior"]],
            csys=fem.lcsys[csys_ref],
        )
    return cons


def get_beam_sections_from_inp(bulk_str, fem):
    """

    https://abaqus-docs.mit.edu/2017/English/SIMACAEELMRefMap/simaelm-c-beamcrosssectlib.htm

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    """
    from ada import Section
    from ada.sections import GeneralProperties

    ass = fem.parent.get_assembly()
    if bulk_str.lower().find("*beam section") == -1:
        return []
    re_beam = re.compile(
        r"(?:\*\*\s*Section:\s*(?P<sec_name>.*?)\s*Profile:\s*(?P<profile_name>.*?)\n|)"
        r"\*Beam Section,\s*elset=(?P<elset>.*?)\s*,\s*material=(?P<material>.*?)\s*,\s*"
        r"(?:temperature=(?P<temperature>.*?),|)\s*(?:section=|sect=)(?P<sec_type>.*?)\n"
        r"(?P<line1>.*?)\n(?P<line2>.*?)$",
        _re_in,
    )

    def interpret_section(profile_name, sec_type, props):
        props_clean = [roundoff(x) for x in filter(lambda x: x.strip() != "", props.split(","))]
        if sec_type.upper() == "BOX":
            b, h, t1, t2, t3, t4 = props_clean
            return Section(
                profile_name,
                "BG",
                h=h,
                w_btn=b,
                w_top=b,
                t_w=t1,
                t_fbtn=t4,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "CIRC":
            return Section(profile_name, "CIRC", r=props_clean[0], parent=fem)
        elif sec_type.upper() == "I":
            (
                l,
                h,
                b1,
                b2,
                t1,
                t2,
                t3,
            ) = props_clean
            return Section(
                profile_name,
                "IG",
                h=h,
                w_btn=b1,
                w_top=b2,
                t_w=t3,
                t_fbtn=t1,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "L":
            b, h, t1, t2 = props_clean
            return Section(profile_name, "HP", h=h, w_btn=b, t_w=t2, t_fbtn=t1, parent=fem)
        elif sec_type.upper() == "PIPE":
            r, t = props_clean
            return Section(profile_name, "TUB", r=r, wt=t, parent=fem)
        elif sec_type.upper() == "TRAPEZOID":
            # Currently converts Trapezoid to general beam
            b, h, a, d = props_clean
            # Assuming the Abaqus trapezoid element is symmetrical
            c = (b - a) / 2

            # The properties were quickly copied from a resource online. Most likely it contains error
            # https: // www.efunda.com / math / areas / trapezoidJz.cfm
            genprops = GeneralProperties(
                Ax=h * (a + b) / 2,
                Ix=h
                * (
                    b * h ** 2
                    + 3 * a * h ** 2
                    + a ** 3
                    + 3 * a * c ** 2
                    + 3 * c * a ** 2
                    + b ** 3
                    + c * b ** 2
                    + a * b ** 2
                    + b * c ** 2
                    + 2 * a * b * c
                    + b * a ** 2
                ),
                Iy=(h ** 3) * (3 * a + b) / 12,
                Iz=h
                * (
                    a ** 3
                    + 3 * a * c ** 2
                    + 3 * c * a ** 2
                    + b ** 3
                    + c * b ** 2
                    + a * b ** 2
                    + 2 * a * b * c
                    + b * a ** 2
                )
                / 12,
            )
            return Section(profile_name, "GENBEAM", genprops=genprops, parent=fem)
        else:
            raise ValueError(f'Currently unsupported section type "{sec_type}"')

    def grab_beam(match):
        d = match.groupdict()
        elset = fem.elsets[d["elset"]]
        name = d["sec_name"] if d["sec_name"] is not None else elset.name
        profile = d["profile_name"] if d["profile_name"] is not None else elset.name
        material = ass.materials.get_by_name(d["material"])
        # material = parent.parent.materials.get_by_name(d['material'])
        temperature = d["temperature"]
        section_type = d["sec_type"]
        geo_props = d["line1"]
        sec = interpret_section(profile, section_type, geo_props)
        beam_y = [float(x.strip()) for x in d["line2"].split(",") if x.strip() != ""]
        metadata = dict(
            temperature=temperature,
            profile=profile.strip(),
            section_type=section_type,
            line1=geo_props,
        )
        res = fem.parent.sections.add(sec)
        if res is not None:
            sec = res
        return FemSection(
            name.strip(),
            sec_type=ElemType.LINE,
            elset=elset,
            section=sec,
            local_y=beam_y,
            material=material,
            metadata=metadata,
            parent=fem,
        )

    return map(grab_beam, re_beam.finditer(bulk_str))


def get_solid_sections_from_inp(bulk_str, fem: FEM):
    secnames = Counter(1, "solidsec")
    a = fem.parent.get_assembly()
    if bulk_str.lower().find("*solid section") == -1:
        return []
    re_solid = re.compile(
        r"(?:\*\s*Section:\s*(.*?)\n|)\*\Solid Section,\s*elset=(.*?)\s*,\s*material=(.*?)\s*$",
        _re_in,
    )
    solid_iter = re_solid.finditer(bulk_str)

    def grab_solid(m_in):
        name = m_in.group(1) if m_in.group(1) is not None else next(secnames)
        elset = m_in.group(2)
        material = m_in.group(3)
        mat = a.materials.get_by_name(material)
        return FemSection(
            name=name,
            sec_type=ElemType.SOLID,
            elset=elset,
            material=mat,
            parent=fem,
        )

    return map(grab_solid, solid_iter)


def get_shell_sections_from_inp(bulk_str, fem: FEM) -> Iterable[FemSection]:
    from ada.fem.elements import Eccentricity

    sh_name = Counter(1, "sh")

    if bulk_str.lower().find("*shell section") == -1:
        return []
    re_offset = r"(?:, offset=(?P<offset>.*?)|)"
    re_controls = r"(?:, controls=(?P<controls>.*?)|)"
    a = fem.parent.get_assembly()
    re_shell = re.compile(
        rf"(?:\*\s*Section:\s*(?P<name>.*?)\n|)\*\Shell Section, elset"
        rf"=(?P<elset>.*?)\s*, material=(?P<material>.*?){re_offset}{re_controls}\s*\n(?P<t>.*?),"
        rf"(?P<int_points>.*?)$",
        _re_in,
    )

    def grab_shell(m):
        d = m.groupdict()
        name = d["name"] if d["name"] is not None else next(sh_name)
        elset = fem.sets.get_elset_from_name(d["elset"])

        material = d["material"]
        mat = a.materials.get_by_name(material)
        thickness = float(d["t"])
        offset = d["offset"]
        if offset is not None:
            # TODO: update this with the latest eccentricity class
            logging.warning("Offset for Shell elements is not yet evaluated")
            for el in elset.members:
                el.eccentricity = Eccentricity(sh_ecc_vector=offset)
        int_points = d["int_points"]
        metadata = dict(controls=d["controls"])
        return FemSection(
            name=name,
            sec_type=ElemType.SHELL,
            thickness=thickness,
            elset=elset,
            material=mat,
            int_points=int_points,
            parent=fem,
            metadata=metadata,
        )

    return map(grab_shell, re_shell.finditer(bulk_str))


def add_interactions_from_bulk_str(bulk_str, assembly: Assembly) -> None:
    gen_name = Counter(1, "general")

    if bulk_str.find("** Interaction") == -1 and bulk_str.find("*Contact") == -1:
        return

    def resolve_surface_ref(surf_ref):
        surf_name = surf_ref.split(".")[-1] if "." in surf_ref else surf_ref
        surf = None
        if surf_name in assembly.fem.surfaces.keys():
            surf = assembly.fem.surfaces[surf_name]

        for p in assembly.get_all_parts_in_assembly():
            if surf_name in p.fem.surfaces.keys():
                surf = p.fem.surfaces[surf_name]
        if surf is None:
            raise ValueError("Unable to find surfaces in assembly parts")

        return surf

    for m in AbaCards.contact_pairs.regex.finditer(bulk_str):
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        surf1 = resolve_surface_ref(d["surf1"])
        surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(Interaction(d["name"], ContactTypes.SURFACE, surf1, surf2, intprop, metadata=d))

    for m in AbaCards.contact_general.regex.finditer(bulk_str):
        s = m.start()
        e = m.endpos
        interact_str = bulk_str[s:e]
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        # surf1 = resolve_surface_ref(d["surf1"])
        # surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(
            Interaction(next(gen_name), "general", None, None, intprop, metadata=dict(aba_bulk=interact_str))
        )


def list_cleanup(membulkstr):
    return membulkstr.replace(",\n", ",").replace("\n", ",")


def is_set_in_part(part: Part, set_name: str, set_type) -> Union[FemSet, Surface]:
    """

    :param part:
    :param set_name:
    :param set_type:

    :return: Set (node, element or surface)
    """

    set_map = {"nset": part.fem.nsets, "elset": part.fem.elsets, "surface": part.fem.surfaces}
    el_map = {"nset": part.fem.nodes, "elset": part.fem.elements}

    if set_name in set_map[set_type].keys():
        return set_map[set_type][set_name]
    else:
        _id = int(set_name)
        return el_map[set_type].from_id(_id)


def grab_set_from_assembly(set_str: str, fem: FEM, set_type) -> Union[FemSet, Surface]:
    res = set_str.split(".")
    if len(res) == 1:
        set_map = {"nset": fem.nsets, "elset": fem.elsets, "surface": fem.surfaces}
        set_name = res[0]
        return set_map[set_type][set_name]
    else:
        set_name = res[1]
        p_name = res[0]
        for part in fem.parent.get_all_parts_in_assembly():
            if p_name == part.fem.instance_name:
                r = is_set_in_part(part, set_name, set_type)
                if r is not None:
                    return r
    raise ValueError(f'No {set_type} "{set_str}" was found')



from itertools import groupby
from typing import List, Union

from ada import FEM
from ada.fem import Bc, Constraint, FemSet
from ada.fem.formats.utils import str_to_int

from . import cards


def get_constraints(bulk_str, fem: FEM) -> List[Constraint]:
    con_map = [m.groupdict() for m in cards.re_bldep.finditer(bulk_str)]
    con_map.sort(key=lambda x: x["master"])
    constraints: List[Constraint] = []
    for m, d in groupby(con_map, key=lambda x: x["master"]):
        constraints += [grab_constraint(m, d, fem)]
    return constraints


def grab_constraint(master, data, fem: FEM):
    m = str_to_int(master)
    m_set = FemSet(f"co{m}_m", [fem.nodes.from_id(m)], "nset")
    slaves = []
    for d in data:
        s = str_to_int(d["slave"])
        slaves.append(fem.nodes.from_id(s))
    s_set = FemSet(f"co{m}_s", slaves, "nset")
    fem.add_set(m_set)
    fem.add_set(s_set)
    return Constraint(f"co{m}", Constraint.TYPES.COUPLING, m_set, s_set, parent=fem)


def get_bcs(bulk_str, fem: FEM) -> List[Bc]:
    return list(filter(lambda x: x is not None, [grab_bc(m, fem) for m in cards.re_bnbcd.finditer(bulk_str)]))


def grab_bc(match, fem: FEM) -> Union[Bc, None]:
    d = match.groupdict()
    node = fem.nodes.from_id(str_to_int(d["nodeno"]))
    for constraint in fem.constraints:
        if node in constraint.m_set.members:
            return None
        if node in constraint.s_set.members:
            return None

    fem_set = fem.sets.add(FemSet(f"bc{node.id}_set", [node], "nset"))
    dofs = []
    for i, c in enumerate(d["content"].replace("\n", "").split()):
        bc_sestype = str_to_int(c.strip())
        if bc_sestype in [0, 4]:
            continue
        dofs.append(i + 1)
    bc = Bc(f"bc{node.id}", fem_set, dofs, parent=fem)
    node.bc = bc
    return bc



import logging

from ada import FEM
from ada.fem.loads import Load

from .write_utils import write_ff


def loads_str(fem: FEM) -> str:
    if len(fem.steps) == 0:
        return ""
    step = fem.steps[0]
    if len(step.loads) == 0:
        return ""

    if len(step.load_cases.keys()) > 0:
        out_str = ""
        for i, lc in enumerate(step.load_cases.values(), start=1):
            out_str += write_ff("TDLOAD", [(4, i, 100 + len(lc.name), 0), (lc.name,)])
            for load in lc.loads:
                out_str += load_str(load, i)
        return out_str

    lid = 1
    load_case_name = "LC1"
    out_str = write_ff("TDLOAD", [(4, lid, 100 + len(load_case_name), 0), (load_case_name,)])
    for load in step.loads:
        out_str += load_str(load, lid)

    return out_str


def load_str(load: Load, lid):
    if load.type in [Load.TYPES.ACC, Load.TYPES.GRAVITY]:
        return load_gravity(load, lid)
    elif load.type == Load.TYPES.FORCE:
        return load_force(load, lid)
    else:
        logging.error(f'Unsupported Load type "{load.type}"')


def load_gravity(load: Load, load_id: int) -> str:
    """Gravity Acceleration field"""
    load_vector = tuple(load.acc_vector)
    return write_ff(
        "BGRAV",
        [(load_id, 0, 0, 0), load_vector],
    )


def load_force(load: Load, load_id: int) -> str:
    """Node with load"""
    lotype = 0
    complx = 0  # Assumed no phase shift
    node_no = load.fem_set.members[0].id
    forces = load.forces
    real_loads_1 = tuple([node_no, 6] + forces[:2])
    real_loads_2 = tuple(forces[2:])
    return write_ff(
        "BNLOAD",
        [(load_id, lotype, complx, 0), real_loads_1, real_loads_2],
    )



import logging
import os
import traceback
from dataclasses import dataclass

import meshio
import numpy as np
from pythreejs import Group

from ada.concepts.levels import FEM
from ada.fem.shapes import ElemShape
from ada.fem.shapes.mesh_types import meshio_to_abaqus_type

from .threejs_utils import edges_to_mesh, faces_to_mesh, vertices_to_mesh
from .utils import get_edges_from_fem, get_faces_from_fem, get_vertices_from_fem


@dataclass
class ViewItem:
    fem: FEM
    vertices: np.array
    edges: np.array
    faces: np.array


class BBox:
    max: list
    min: list
    center: list


def fem_to_mesh(
    fem: FEM, face_colors=None, vertex_colors=(8, 8, 8), edge_color=(8, 8, 8), edge_width=1, vertex_width=1
):
    vertices, faces, edges = get_vertices_from_fem(fem), get_faces_from_fem(fem), get_edges_from_fem(fem)

    name = fem.name

    vertices_m = vertices_to_mesh(f"{name}_vertices", vertices, vertex_colors, vertex_width)
    edges_m = edges_to_mesh(f"{name}_edges", vertices, edges, edge_color=edge_color, linewidth=edge_width)
    faces_mesh = faces_to_mesh(f"{name}_faces", vertices, faces, colors=face_colors)

    return vertices_m, edges_m, faces_mesh


class FemRenderer:
    def __init__(self):
        self._view_items = []
        self._meshes = []

        # the group of 3d and 2d objects to render
        self._displayed_pickable_objects = Group()

    def add_fem(self, fem: FEM):
        vertices, faces, edges = get_vertices_from_fem(fem), get_faces_from_fem(fem), get_edges_from_fem(fem)
        self._view_items.append(ViewItem(fem, vertices, edges, faces))

    def to_mesh(self):
        for vt in self._view_items:
            self._view_to_mesh(vt)

    def _view_to_mesh(
        self, vt, face_colors=None, vertex_colors=(8, 8, 8), edge_color=(8, 8, 8), edge_width=1, vertex_width=1
    ):
        """

        :param vt:
        :type vt: ViewItem
        :return:
        """
        fem = vt.fem
        vertices = vt.vertices
        edges = vt.edges
        faces = vt.faces

        vertices_m = vertices_to_mesh(f"{fem.name}_vertices", vertices, vertex_colors, vertex_width)
        edges_m = edges_to_mesh(f"{fem.name}_edges", vertices, edges, edge_color=edge_color, linewidth=edge_width)
        face_geom, faces_m = faces_to_mesh(f"{fem.name}_faces", vertices, faces, colors=face_colors)

        return vertices_m, edges_m, faces_m

    def get_bounding_box(self):
        bounds = np.asarray([get_bounding_box(m) for m in self._meshes], dtype="float32")
        mi, ma = np.min(bounds, 0), np.max(bounds, 0)
        center = (mi + ma) / 2
        return mi, ma, center


def get_edges_and_faces_from_meshio(mesh: meshio.Mesh):
    edges = []
    faces = []
    for cell_block in mesh.cells:
        el_type = meshio_to_abaqus_type[cell_block.type]
        for elem in cell_block.data:
            elem_shape = ElemShape(el_type, elem)
            edges += elem_shape.edges
            if elem_shape.type in elem_shape.TYPES.lines:
                continue
            faces += elem_shape.faces
    return edges, faces


def get_faces_for_bm_elem(elem):
    """

    :param elem:
    :type elem: ada.fem.Elem
    :return:
    """

    # if ElemShapes.beam


def get_bounding_box(vertices):
    return np.min(vertices, 0), np.max(vertices, 0)


def magnitude(u):
    return np.sqrt(u[0] ** 2 + u[1] ** 2 + u[2] ** 2)


def visualize_it(res_file, temp_dir=".temp", default_index=0):
    import pathlib

    import meshio
    from ipygany import ColorBar, IsoColor, PolyMesh, Scene, Warp, colormaps
    from IPython.display import clear_output, display
    from ipywidgets import AppLayout, Dropdown, FloatSlider, VBox, jslink

    from ada.core.utils import vector_length

    res_file = pathlib.Path(res_file).resolve().absolute()
    suffix = res_file.suffix.lower()

    suffix_map = {".rmed": "med", ".vtu": None}

    imesh = meshio.read(res_file, file_format=suffix_map[suffix])
    imesh.point_data = {key.replace(" ", "_"): value for key, value in imesh.point_data.items()}

    def filter_keys(var):
        if suffix == ".vtu" and var != "U":
            return False
        if suffix == ".rmed" and var == "point_tags":
            return False
        return True

    warp_data = [key for key in filter(filter_keys, imesh.point_data.keys())]
    magn_data = []
    for d in warp_data:
        res = [vector_length(v[:3]) for v in imesh.point_data[d]]
        res_norm = [r / max(res) for r in res]
        magn_data_name = f"{d}_magn"
        imesh.point_data[magn_data_name] = np.array(res_norm, dtype=np.float64)
        magn_data.append(magn_data_name)

    imesh.field_data = {key: np.array(value) for key, value in imesh.field_data.items()}

    tf = (pathlib.Path(temp_dir).resolve().absolute() / res_file.name).with_suffix(".vtu")

    if tf.exists():
        os.remove(tf)
    os.makedirs(tf.parent, exist_ok=True)
    imesh.write(tf)

    mesh = PolyMesh.from_vtk(str(tf))
    mesh.default_color = "gray"

    warp_vec = warp_data[default_index]
    try:
        colored_mesh = IsoColor(mesh, input=magn_data[default_index], min=0.0, max=1.0)
    except KeyError as e:
        trace_str = traceback.format_exc()
        logging.error(f'KeyError "{e}"\nTrace: "{trace_str}"')
        colored_mesh = mesh
    except ImportError as e:
        trace_str = traceback.format_exc()
        logging.error("This might be")
        logging.error(f'ImportError "{e}"\nTrace: "{trace_str}"')
        return

    warped_mesh = Warp(colored_mesh, input=warp_vec, warp_factor=0.0)

    warp_slider = FloatSlider(value=0.0, min=-1.0, max=1.0)

    jslink((warped_mesh, "factor"), (warp_slider, "value"))

    # Create a colorbar widget
    colorbar = ColorBar(colored_mesh)

    # Colormap choice widget
    colormap = Dropdown(options=colormaps, description="colormap:")

    jslink((colored_mesh, "colormap"), (colormap, "index"))

    # EigenValue choice widget
    eig_map = Dropdown(options=warp_data, description="Data Value:")

    scene = Scene([warped_mesh])
    app = AppLayout(
        left_sidebar=scene, right_sidebar=VBox((eig_map, warp_slider, colormap, colorbar)), pane_widths=[2, 0, 1]
    )

    def change_input(change):
        vec_name = change["new"]
        logging.info(vec_name)
        colored_mesh.input = vec_name + "_magn"
        warped_mesh.input = vec_name
        # Highly inefficient but likely needed due to bug https://github.com/QuantStack/ipygany/issues/69
        clear_output()
        display(app)

    eig_map.observe(change_input, names=["value"])

    return app



from __future__ import annotations

import logging
from typing import Callable, Dict

from . import abaqus, calculix, code_aster, sesam, usfos
from .utils import interpret_fem


class FEATypes:
    CODE_ASTER = "code_aster"
    CALCULIX = "calculix"
    ABAQUS = "abaqus"
    SESAM = "sesam"
    USFOS = "usfos"

    all = [CODE_ASTER, CALCULIX, ABAQUS, SESAM, USFOS]


fem_imports = {
    FEATypes.ABAQUS: abaqus.read_fem,
    FEATypes.SESAM: sesam.read_fem,
    FEATypes.CODE_ASTER: code_aster.read_fem,
}

fem_exports = {
    FEATypes.ABAQUS: abaqus.to_fem,
    FEATypes.CALCULIX: calculix.to_fem,
    FEATypes.CODE_ASTER: code_aster.to_fem,
    FEATypes.SESAM: sesam.to_fem,
    FEATypes.USFOS: usfos.to_fem,
}

fem_executables: Dict[str, Callable] = {
    FEATypes.ABAQUS: abaqus.run_abaqus,
    FEATypes.CALCULIX: calculix.run_calculix,
    FEATypes.CODE_ASTER: code_aster.run_code_aster,
    FEATypes.SESAM: sesam.run_sesam,
}

fem_solver_map = {FEATypes.SESAM: "sestra", FEATypes.CALCULIX: "ccx"}


class FemConverters:
    DEFAULT = "default"
    MESHIO = "meshio"


def get_fem_converters(fem_file, fem_format, fem_converter):
    from ada.fem.formats.mesh_io import meshio_read_fem, meshio_to_fem

    if fem_format is None:
        fem_format = interpret_fem(fem_file)

    if fem_converter == FemConverters.DEFAULT:
        fem_importer = fem_imports.get(fem_format, None)
        fem_exporter = fem_exports.get(fem_format, None)
    elif fem_converter.lower() == FemConverters.MESHIO:
        fem_importer = meshio_read_fem
        fem_exporter = meshio_to_fem
    else:
        raise ValueError(f'Unrecognized fem_converter "{fem_converter}". Only "meshio" and "default" are supported')

    return fem_importer, fem_exporter


def export_fem(assembly, name, analysis_dir, fem_format, fem_converter, metadata):
    _, fem_exporter = get_fem_converters("", fem_format, fem_converter)
    metadata = dict() if metadata is None else metadata
    metadata["fem_format"] = fem_format
    try:
        fem_exporter(assembly, name, analysis_dir, metadata)
        return True
    except IOError as e:
        logging.error(e)
        return False



from __future__ import annotations

from typing import List, Union

from ada.concepts.points import Node

from .common import FemBase
from .elements import Elem


class SetTypes:
    NSET = "nset"
    ELSET = "elset"

    all = [NSET, ELSET]


class FemSet(FemBase):
    """

    :param name: Name of Set
    :param members: Set Members
    :param set_type: Type of set (either 'nset' or 'elset')
    :param metadata: Metadata for object
    :param parent: Parent object
    """

    TYPES = SetTypes

    def __init__(self, name, members: Union[None, List[Union[Elem, Node]]], set_type=None, metadata=None, parent=None):
        super().__init__(name, metadata, parent)
        if set_type is None:
            set_type = eval_set_type_from_members(members)

        self._set_type = set_type
        if self.type not in SetTypes.all:
            raise ValueError(f'set type "{set_type}" is not valid')
        self._members = members

    def __len__(self):
        return len(self._members)

    def __contains__(self, item):
        return item.id in self._members

    def __getitem__(self, index):
        return self._members[index]

    def __add__(self, other: FemSet) -> FemSet:
        self.add_members(other.members)
        return self

    def add_members(self, members: List[Union[Elem, Node]]):
        self._members += members

    @property
    def type(self):
        return self._set_type.lower()

    @property
    def members(self) -> List[Union[Elem, Node]]:
        return self._members

    def __repr__(self):
        return f'FemSet({self.name}, type: "{self.type}", members: "{len(self.members)}")'


def eval_set_type_from_members(members: List[Union[Elem, Node]]):
    res = set([type(mem) for mem in members])
    if len(res) == 1 and type(members[0]) is Node:
        return "nset"
    elif len(res) == 1 and type(members[0]) is Elem:
        return "elset"
    else:
        raise ValueError("Currently Mixed Femsets are not allowed")
        # return "mixed"



from __future__ import annotations

import logging
from typing import List, Union

import numpy as np

from ada.base.physical_objects import BackendGeom
from ada.config import Settings as _Settings
from ada.core.utils import (
    Counter,
    angle_between,
    calc_zvec,
    roundoff,
    unit_vector,
    vector_length,
)
from ada.ifc.utils import create_guid
from ada.materials.utils import get_material
from ada.sections.utils import get_section

from .curves import ArcSegment
from .points import Node


class Pipe(BackendGeom):
    def __init__(
        self,
        name,
        points,
        sec,
        mat="S355",
        content=None,
        metadata=None,
        colour=None,
        units="m",
        guid=None,
        ifc_elem=None,
    ):
        super().__init__(name, guid=guid, metadata=metadata, units=units, ifc_elem=ifc_elem)

        self._section, _ = get_section(sec)
        self._section.parent = self
        self._material = get_material(mat)
        self._content = content
        self.colour = colour

        self._n1 = points[0] if type(points[0]) is Node else Node(points[0], units=units)
        self._n2 = points[-1] if type(points[-1]) is Node else Node(points[-1], units=units)
        self._points = [Node(n, units=units) if type(n) is not Node else n for n in points]
        self._segments = []
        self._build_pipe()

    def _build_pipe(self):
        from ada.core.curve_utils import make_arc_segment

        segs = []
        for p1, p2 in zip(self.points[:-1], self.points[1:]):
            if vector_length(p2.p - p1.p) == 0.0:
                logging.info("skipping zero length segment")
                continue
            segs.append([p1, p2])
        segments = segs

        seg_names = Counter(prefix=self.name + "_")

        # Make elbows and adjust segments
        props = dict(section=self.section, material=self.material, parent=self, units=self.units)
        angle_tol = 1e-1
        len_tol = _Settings.point_tol if self.units == "m" else _Settings.point_tol * 1000
        for i, (seg1, seg2) in enumerate(zip(segments[:-1], segments[1:])):
            p11, p12 = seg1
            p21, p22 = seg2
            vlen1 = vector_length(seg1[1].p - seg1[0].p)
            vlen2 = vector_length(seg2[1].p - seg2[0].p)

            if vlen1 < len_tol or vlen2 == len_tol:
                logging.error(f'Segment Length is below point tolerance for unit "{self.units}". Skipping')
                continue

            xvec1 = unit_vector(p12.p - p11.p)
            xvec2 = unit_vector(p22.p - p21.p)
            a = angle_between(xvec1, xvec2)
            res = True if abs(abs(a) - abs(np.pi)) < angle_tol or abs(abs(a) - 0.0) < angle_tol else False

            if res is True:
                self._segments.append(PipeSegStraight(next(seg_names), p11, p12, **props))
            else:
                if p12 != p21:
                    logging.error("No shared point found")

                if i != 0 and len(self._segments) > 0:
                    pseg = self._segments[-1]
                    prev_p = (pseg.p1.p, pseg.p2.p)
                else:
                    prev_p = (p11.p, p12.p)
                try:
                    seg1, arc, seg2 = make_arc_segment(prev_p[0], prev_p[1], p22.p, self.pipe_bend_radius * 0.99)
                except ValueError as e:
                    logging.error(f"Error: {e}")  # , traceback: "{traceback.format_exc()}"')
                    continue
                except RuntimeError as e:
                    logging.error(f"Error: {e}")  # , traceback: "{traceback.format_exc()}"')
                    continue

                if i == 0 or len(self._segments) == 0:
                    self._segments.append(
                        PipeSegStraight(
                            next(seg_names), Node(seg1.p1, units=self.units), Node(seg1.p2, units=self.units), **props
                        )
                    )
                else:
                    if len(self._segments) == 0:
                        continue
                    pseg = self._segments[-1]
                    pseg.p2 = Node(seg1.p2, units=self.units)

                self._segments.append(
                    PipeSegElbow(
                        next(seg_names) + "_Elbow",
                        Node(seg1.p1, units=self.units),
                        Node(p21.p, units=self.units),
                        Node(seg2.p2, units=self.units),
                        arc.radius,
                        **props,
                        arc_seg=arc,
                    )
                )
                self._segments.append(
                    PipeSegStraight(
                        next(seg_names), Node(seg2.p1, units=self.units), Node(seg2.p2, units=self.units), **props
                    )
                )

    @property
    def segments(self) -> List[Union[PipeSegStraight, PipeSegElbow]]:
        return self._segments

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def points(self):
        return self._points

    @property
    def start(self):
        return self.points[0]

    @property
    def end(self):
        return self.points[-1]

    @property
    def metadata(self):
        return self._metadata

    @property
    def pipe_bend_radius(self):
        wt = self.section.wt
        r = self.section.r
        d = r * 2
        w_tol = 0.125 if self.units == "m" else 125
        cor_tol = 0.003 if self.units == "m" else 3
        corr_t = (wt - (wt * w_tol)) - cor_tol
        d -= 2.0 * corr_t

        return roundoff(d + corr_t / 2.0)

    @property
    def section(self):
        """

        :return:
        :rtype: Section
        """
        return self._section

    @property
    def n1(self):
        return self._n1

    @property
    def n2(self):
        return self._n2

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if value != self._units:
            self.n1.units = value
            self.n2.units = value
            self.section.units = value
            self.material.units = value
            self._segments = []
            for p in self.points:
                p.units = value
            self._build_pipe()
            self._units = value

    def _generate_ifc_pipe(self):
        from ada.core.constants import X, Z
        from ada.ifc.utils import create_local_placement, create_property_set

        if self.parent is None:
            raise ValueError("Cannot build ifc element without parent")

        a = self.get_assembly()
        f = a.ifc_file

        owner_history = a.user.to_ifc()
        parent = self.parent.get_ifc_elem()

        placement = create_local_placement(
            f,
            origin=self.n1.p.astype(float).tolist(),
            loc_x=X,
            loc_z=Z,
            relative_to=parent.ObjectPlacement,
        )

        ifc_elem = f.createIfcSpace(
            self.guid,
            owner_history,
            self.name,
            "Description",
            None,
            placement,
            None,
            None,
            None,
        )

        f.createIfcRelAggregates(
            create_guid(),
            owner_history,
            "Site Container",
            None,
            parent,
            [ifc_elem],
        )
        if len(self.metadata.keys()) > 0:
            props = create_property_set("Properties", f, self.metadata)
            f.createIfcRelDefinesByProperties(
                create_guid(),
                owner_history,
                "Properties",
                None,
                [ifc_elem],
                props,
            )

        return ifc_elem

    def get_ifc_elem(self):
        if self._ifc_elem is None:
            self._ifc_elem = self._generate_ifc_pipe()

            a = self.get_assembly()
            f = a.ifc_file

            owner_history = a.user.to_ifc()

            segments = []
            for param_seg in self._segments:
                if type(param_seg) is PipeSegStraight:
                    assert isinstance(param_seg, PipeSegStraight)
                    res = param_seg.get_ifc_elem()
                else:
                    assert isinstance(param_seg, PipeSegElbow)
                    res = param_seg.get_ifc_elem()
                if res is None:
                    logging.error(f'Branch "{param_seg.name}" was not converted to ifc element')
                f.add(res)
                segments += [res]

            f.createIfcRelContainedInSpatialStructure(
                create_guid(),
                owner_history,
                "Pipe Segments",
                None,
                segments,
                self._ifc_elem,
            )

        return self._ifc_elem

    def __repr__(self):
        return f"Pipe({self.name}, {self.section})"


class PipeSegStraight(BackendGeom):
    def __init__(
        self,
        name,
        p1,
        p2,
        section,
        material,
        parent=None,
        guid=None,
        metadata=None,
        units="m",
        colour=None,
        ifc_elem=None,
    ):
        super(PipeSegStraight, self).__init__(name, guid, metadata, units, parent, colour, ifc_elem=ifc_elem)
        self.p1 = p1
        self.p2 = p2
        self.section = section
        self.material = material

    @property
    def xvec1(self):
        return self.p2.p - self.p1.p

    @property
    def zvec(self):
        return calc_zvec(self.xvec1)

    @property
    def line(self):
        from ada.occ.utils import make_edge

        return make_edge(self.p1, self.p2)

    @property
    def shell(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        return sweep_pipe(self.line, self.xvec1, self.section.r, self.section.wt, ElemType.SHELL)

    @property
    def solid(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        return sweep_pipe(self.line, self.xvec1, self.section.r, self.section.wt, ElemType.SOLID)

    def _generate_ifc_elem(self):
        from ada.core.constants import O, X, Z
        from ada.ifc.utils import (  # create_ifcrevolveareasolid,
            create_ifc_placement,
            create_ifcpolyline,
            to_real,
        )

        if self.parent is None:
            raise ValueError("Parent cannot be None for IFC export")

        a = self.parent.get_assembly()
        f = a.ifc_file

        context = f.by_type("IfcGeometricRepresentationContext")[0]
        owner_history = a.user.to_ifc()

        p1 = self.p1
        p2 = self.p2

        ifcdir = f.createIfcDirection((0.0, 0.0, 1.0))

        rp1 = to_real(p1.p)
        rp2 = to_real(p2.p)
        xvec = unit_vector(p2.p - p1.p)
        a = angle_between(xvec, np.array([0, 0, 1]))
        zvec = np.array([0, 0, 1]) if a != np.pi and a != 0 else np.array([1, 0, 0])
        yvec = unit_vector(np.cross(zvec, xvec))
        seg_l = vector_length(p2.p - p1.p)

        extrusion_placement = create_ifc_placement(f, (0.0, 0.0, 0.0), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0))

        solid = f.createIfcExtrudedAreaSolid(self.section.ifc_profile, extrusion_placement, ifcdir, seg_l)

        polyline = create_ifcpolyline(f, [rp1, rp2])

        axis_representation = f.createIfcShapeRepresentation(context, "Axis", "Curve3D", [polyline])
        body_representation = f.createIfcShapeRepresentation(context, "Body", "SweptSolid", [solid])

        product_shape = f.createIfcProductDefinitionShape(None, None, [axis_representation, body_representation])

        origin = f.createIfcCartesianPoint(O)
        local_z = f.createIfcDirection(Z)
        local_x = f.createIfcDirection(X)
        d237 = f.createIfcLocalPlacement(None, f.createIfcAxis2Placement3D(origin, local_z, local_x))

        d256 = f.createIfcCartesianPoint(rp1)
        d257 = f.createIfcDirection(to_real(xvec))
        d258 = f.createIfcDirection(to_real(yvec))
        d236 = f.createIfcAxis2Placement3D(d256, d257, d258)
        local_placement = f.createIfcLocalPlacement(d237, d236)

        pipe_segment = f.createIfcPipeSegment(
            create_guid(),
            owner_history,
            self.name,
            "An awesome pipe",
            None,
            local_placement,
            product_shape,
            None,
        )

        ifc_mat = self.material.ifc_mat
        mat_profile = f.createIfcMaterialProfile(
            self.material.name, None, ifc_mat, self.section.ifc_profile, None, None
        )
        mat_profile_set = f.createIfcMaterialProfileSet(None, None, [mat_profile], None)
        mat_profile_set = f.createIfcMaterialProfileSetUsage(mat_profile_set, 8, None)
        f.createIfcRelAssociatesMaterial(create_guid(), None, None, None, [pipe_segment], mat_profile_set)

        return pipe_segment


class PipeSegElbow(BackendGeom):
    def __init__(
        self,
        name,
        p1,
        p2,
        p3,
        bend_radius,
        section,
        material,
        parent=None,
        guid=None,
        metadata=None,
        units="m",
        colour=None,
        arc_seg=None,
    ):
        super(PipeSegElbow, self).__init__(name, guid, metadata, units, parent, colour)
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.bend_radius = bend_radius
        self.section = section
        self.material = material
        self._arc_seg = arc_seg

    @property
    def parent(self) -> Pipe:
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def xvec1(self):
        return self.p2.p - self.p1.p

    @property
    def xvec2(self):
        return self.p3.p - self.p2.p

    @property
    def zvec(self):
        return calc_zvec(self.xvec1)

    @property
    def line(self):
        from ada.core.curve_utils import make_edges_and_fillet_from_3points

        if self.arc_seg.edge_geom is None:
            _, _, fillet = make_edges_and_fillet_from_3points(self.p1, self.p2, self.p3, self.bend_radius)
            edge = fillet
        else:
            edge = self.arc_seg.edge_geom
        return edge

    @property
    def shell(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        i = self.parent.segments.index(self)
        if i != 0:
            pseg = self.parent.segments[i - 1]
            xvec = pseg.xvec1
        else:
            xvec = self.xvec1

        return sweep_pipe(self.line, xvec, self.section.r, self.section.wt, ElemType.SHELL)

    @property
    def solid(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        i = self.parent.segments.index(self)
        if i != 0:
            pseg = self.parent.segments[i - 1]
            xvec = pseg.xvec1
        else:
            xvec = self.xvec1

        return sweep_pipe(self.line, xvec, self.section.r, self.section.wt, ElemType.SOLID)

    @property
    def arc_seg(self) -> ArcSegment:
        return self._arc_seg

    def _elbow_tesselated(self, f, schema, a):
        from ada.ifc.utils import get_tolerance, tesselate_shape

        shape = self.solid

        if shape is None:
            logging.error(f"Unable to create geometry for Branch {self.name}")
            return None

        serialized_geom = tesselate_shape(shape, schema, get_tolerance(a.units))
        ifc_shape = f.add(serialized_geom)

        return ifc_shape

    def _elbow_revolved_solid(self, f, context):
        from ada.core.constants import O, X, Z
        from ada.core.curve_utils import get_center_from_3_points_and_radius
        from ada.core.utils import normal_to_points_in_plane
        from ada.ifc.utils import create_ifc_placement

        center, _, _, _ = get_center_from_3_points_and_radius(self.p1.p, self.p2.p, self.p3.p, self.bend_radius)

        opening_axis_placement = create_ifc_placement(f, O, Z, X)

        profile = self.section.ifc_profile
        normal = normal_to_points_in_plane([self.arc_seg.p1, self.arc_seg.p2, self.arc_seg.midpoint])
        revolve_axis = self.arc_seg.center + normal
        revolve_angle = 10

        ifcorigin = f.createIfcCartesianPoint(self.arc_seg.p1.astype(float).tolist())
        ifcaxis1dir = f.createIfcAxis1Placement(ifcorigin, f.createIfcDirection(revolve_axis.astype(float).tolist()))

        ifc_shape = f.createIfcRevolvedAreaSolid(profile, opening_axis_placement, ifcaxis1dir, revolve_angle)

        curve = f.createIfcTrimmedCurve()

        body = f.createIfcShapeRepresentation(context, "Body", "SweptSolid", [ifc_shape])
        axis = f.createIfcShapeRepresentation(context, "Axis", "Curve3D", [curve])
        prod_def_shp = f.createIfcProductDefinitionShape(None, None, (axis, body))

        return prod_def_shp

    def _generate_ifc_elem(self):
        from ada.ifc.utils import create_local_placement

        if self.parent is None:
            raise ValueError("Parent cannot be None for IFC export")

        a = self.parent.get_assembly()
        f = a.ifc_file

        context = f.by_type("IfcGeometricRepresentationContext")[0]
        owner_history = a.user.to_ifc()
        schema = a.ifc_file.wrapped_data.schema

        if _Settings.make_param_elbows is False:
            ifc_elbow = self._elbow_tesselated(f, schema, a)
            # Link to representation context
            for rep in ifc_elbow.Representations:
                rep.ContextOfItems = context
        else:
            ifc_elbow = self._elbow_revolved_solid(f, context)

        pfitting_placement = create_local_placement(f)

        pfitting = f.createIfcPipeFitting(
            create_guid(),
            owner_history,
            self.name,
            "An awesome Elbow",
            None,
            pfitting_placement,
            ifc_elbow,
            None,
            None,
        )

        ifc_mat = self.material.ifc_mat
        mat_profile = f.createIfcMaterialProfile(
            self.material.name, None, ifc_mat, self.section.ifc_profile, None, None
        )
        mat_profile_set = f.createIfcMaterialProfileSet(None, None, [mat_profile], None)
        mat_profile_set = f.createIfcMaterialProfileSetUsage(mat_profile_set, 8, None)
        f.createIfcRelAssociatesMaterial(create_guid(), None, None, None, [pfitting], mat_profile_set)

        return pfitting



import pathlib

from ..utils import LocalExecute


def run_calculix(
    inp_path, cpus=2, gpus=None, run_ext=False, metadata=None, execute=True, exit_on_complete=True, run_in_shell=False
):
    inp_path = pathlib.Path(inp_path)

    ccx = CalculixExecute(
        inp_path,
        cpus=cpus,
        gpus=gpus,
        run_ext=run_ext,
        metadata=metadata,
        auto_execute=execute,
        run_in_shell=run_in_shell,
    )
    return ccx.run(exit_on_complete=exit_on_complete)


class CalculixExecute(LocalExecute):
    def run(self, exit_on_complete=True):
        from ada.fem.formats import FEATypes

        exe_path = self.get_exe(FEATypes.CALCULIX)
        out = self._run_local(f"{exe_path} -i {self.analysis_name}", exit_on_complete=exit_on_complete)
        return out



import datetime
from typing import Union

from ada.core.utils import get_current_user
from ada.fem.steps import Step, StepEigen, StepImplicit

from .templates import sestra_eig_inp_str, sestra_header_inp_str, sestra_static_inp_str


def write_sestra_inp(name, step: Union[StepEigen]):
    step_map = {Step.TYPES.EIGEN: write_sestra_eig_str, Step.TYPES.STATIC: write_sestra_static_str}
    step_str_writer = step_map.get(step.type, None)
    if step_str_writer is None:
        raise ValueError(f'Step type "{step.type}" is not supported yet for Ada-Sestra ')

    now = datetime.datetime.now()
    date_str = now.strftime("%d-%b-%Y")
    clock_str = now.strftime("%H:%M:%S")
    user = get_current_user()
    head_str = sestra_header_inp_str.format(date_str=date_str, clock_str=clock_str, user=user)
    return head_str + step_str_writer(name, step)


def write_sestra_eig_str(name: str, step: StepEigen):
    return sestra_eig_inp_str.format(name=name, modes=step.num_eigen_modes, supnr=1)


def write_sestra_static_str(name: str, step: StepImplicit):
    return sestra_static_inp_str.format(name=name, supnr=1)



import logging

from ada import Section
from ada.sections import SectionCat

from .write_utils import write_ff


def general_beam(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GBEAMG",
        [
            (sec_id, 0, p.Ax, p.Ix),
            (p.Iy, p.Iz, p.Iyz, p.Wxmin),
            (p.Wymin, p.Wzmin, p.Shary, p.Sharz),
            (p.Shceny, p.Shcenz, p.Sy, p.Sz),
        ],
    )


def angular(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GLSEC",
        [
            (sec_id, sec.h, sec.t_w, sec.w_btn),
            (sec.t_fbtn, p.Sfy, p.Sfz, 1),
        ],
    )


def box(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GBOX",
        [
            (sec_id, sec.h, sec.t_w, sec.t_fbtn),
            (sec.t_ftop, sec.w_btn, p.Sfy, p.Sfz),
        ],
    )


def iprofile(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GIORH",
        [
            (sec_id, sec.h, sec.t_w, sec.w_top),
            (sec.t_ftop, sec.w_btn, sec.t_fbtn, p.Sfy),
            (p.Sfz,),
        ],
    )


def tubular(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GPIPE",
        [(sec_id, (sec.r - sec.wt) * 2, sec.r * 2, sec.wt), (p.Sfy, p.Sfz)],
    )


def circular(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff(
        "GPIPE",
        [(sec_id, (sec.r - sec.r * 0.99) * 2, sec.r * 2, sec.wt), (p.Sfy, p.Sfz)],
    )


def flatbar(sec: Section, sec_id) -> str:
    p = sec.properties
    return write_ff("GBARM", [(sec_id, sec.h, sec.w_top, sec.w_btn), (p.Sfy, p.Sfz)])


def write_bm_section(sec: Section, sec_id: int) -> str:
    bt = SectionCat.BASETYPES
    base_type = SectionCat.get_shape_type(sec)
    sec_map = {
        bt.ANGULAR: angular,
        bt.BOX: box,
        bt.IPROFILE: iprofile,
        bt.TPROFILE: iprofile,
        bt.TUBULAR: tubular,
        bt.CIRCULAR: circular,
        bt.FLATBAR: flatbar,
    }

    sec_str = general_beam(sec, sec_id)

    sec_str_writer = sec_map.get(base_type, None)
    if base_type == bt.GENERAL:
        return sec_str

    if sec_str_writer is None:
        logging.error(f'Unable to convert "{sec}". This will be exported as general section only')

    sec_str += sec_str_writer(sec, sec_id)

    return sec_str
